<?php

	class in8_order_items extends WC_API_Resource {


		protected $base = '/order_items';
		public function register_routes( $sroutes ) {

			# GET /Update Order Item details and recalculate.

			$sroutes[ $this->base ] = [
				[ [ $this, 'order_items_update' ], WC_API_Server::READABLE | WC_API_Server::EDITABLE | WC_API_Server::ACCEPT_DATA ],
			];

			return $sroutes;
		}

		public function order_items_update() {
			
			error_reporting(-1);
			// They are looking up data
			$ordId        = $_GET["id"];

			$data = $_REQUEST['data'];
			$data = json_decode( base64_decode( $data ), TRUE );
			$lines = $data["lines"];


			if ( empty($ordId) ) {
				echo json_encode( [ "errors" => "Missing id argument. POST" ] );
				exit;
			}

			# check if the lines are set
			if( empty($lines) ){
				echo json_encode( [ "errors" => "You're not sending lines to save. POST" ] );
				exit;
			}

			# get order
			$order = new WC_Order($ordId);
			/* @var WC_Order $order */
			if( empty($order->post) || $order->post->post_type != 'shop_order' ){
				echo json_encode( [ "errors" => "Order doesn't exist. POST" ] );
				exit;
			}

			$acceptedLineTypes = [
				'line_item',
//				'shipping',
				'fee',
				'coupon',
			];
			foreach($acceptedLineTypes as $type){
				foreach($order->get_items($type) as $id => $item){
					wc_delete_order_item( $id );
				}
			}

			$lines = apply_filters('in8_endpoint_order_items_lines', $lines, $order);
			foreach($lines as $line) {
				if( !isset($line['line_type']) ){
					$type = 'line_item';
				}else{
					if( !in_array($line['line_type'], $acceptedLineTypes) ){
						# we don't accept this line type (should we kill the process?)
						continue;
					}
					$type = $line['line_type'];
				}

				$helper = new WC_API_Orders_Helper();
				$helper->set_item($order, $type, $line, 'create');

			}

			// calculate totals and set them
			$order->calculate_totals();
			
			echo json_encode( [ "success" => "Items saved" ] );
			exit;
		}

	}

	class WC_API_Orders_Helper extends WC_API_Orders{
		public function __construct(){
			
		}
		public function set_item( $order, $item_type, $itemData, $action ) {
			$genericItemId = 0;
			if( class_exists('In8_Customers_Sales_Options') ){
				$genericItemId = In8_Customers_Sales_Options::getGenericItem();
			}
			
			if( $item_type == 'line_item' && (!empty($itemData['product_id']) && $genericItemId == $itemData['product_id']) ){
				$item = new WC_Order_Item_Product();
				if( isset($itemData['id']) ){
					$item = $order->get_item($itemData['id']);
				}
				$item->set_props( $itemData );
				
				$product = wc_get_product($itemData['product_id']);
				$item->set_product($product);
				$item->set_order_id( $order->get_id() );
				$item->set_props( $itemData );

				// save product info on the item when it is a generic item
				$genericItemTitle = !empty($itemData['description']) ? $itemData['description'] : $itemData['displayname'];
				if( !empty($genericItemTitle) ) {
					$item->set_name($genericItemTitle);
				}
				$item->add_meta_data('internalid', $itemData['internalid']);
				$item->add_meta_data('sku', $itemData['_sku']);
				
				$item = apply_filters('in8_order_save_generic_item', $item, $itemData, $order->get_id());
				$item->save();
				
				if( !isset($itemData['id']) ){
					$order->add_item($item);
				}
			}else {
				parent::set_item( $order, $item_type, $itemData, $action );
			}
		}
	}

