<?php

    class In8_SalesOrder {

        public $in8_salesorder_logger;
        public static $optionsTag = 'In8_SalesOrder'; //plugin tag name
        public static $optionsTagCM = 'In8_SalesOrder_CM'; //conditional mapping
        public static $optionsTagCO = 'In8_SalesOrder_CO'; //configurable options
        public static $optionsTagFP = 'In8_SalesOrder_FP'; //forms, payment methods, and shipping methods from NS
        private static $salesOrderFieldMapping = [
            'paymentmethod_AmericanExpress' => 0,
            'paymentmethod_Discover' => 0,
            'paymentmethod_Visa' => 0,
            'paymentmethod_MasterCard' => 0,
            'paymentmethod_JCB' => 0,
            'paymentmethod_DinersClub' => 0,
            'paymentmethod_Paypal' => 0,
            'currency_EUR' => 0,
            'currency_USD' => 0,
            'currency_GBP' => 0,
            'customform' => 0,
            'terms' => 0,
            'location' => 0,
            'leadsource' => 0,
            'salesrep' => 0,
            'partner' => 0,
            'istaxable' => 'T',
            'shipcomplete' => 'F',
            'subsidiary' => 0,
            'department' => 0,
            'opportunity' => 0,
            'class' => 0,
            'orderstatus' => 0,
            'couponcode' => 0,
            'shippingtaxcode' => 0,
        ];
        private static $salesOrderFieldHelpText = [
            'paymentmethod_AmericanExpress' => "NetSuite internal ID for American Express payment method",
            'paymentmethod_Discover' => "NetSuite internal ID for Discover payment method",
            'paymentmethod_Visa' => "NetSuite internal ID for Visa payment method",
            'paymentmethod_MasterCard' => "NetSuite internal ID for Mastercard payment method",
            'paymentmethod_Paypal' => "NetSuite internal ID for Paypal payment method",
            'currency_EUR' => "NetSuite currency internal ID for EUR",
            'currency_USD' => "NetSuite currency internal ID for USD",
            'currency_GBP' => "NetSuite currency internal ID for GBP",
            'customform' => "NetSuite order Custom form internal ID",
            'terms' => "NetSuite order terms internal ID",
            'location' => "NetSuite order location internal ID",
            'leadsource' => "NetSuite order leadsource NetSuite internal ID",
            'salesrep' => "NetSuite order salesrep internal ID",
            'partner' => "Order partner NetSuite internal ID",
            'istaxable' => "T or F, if true order taxes will be sent to NetSuite",
            'shipcomplete' => "T or F, set true if you want to ship this order when there are enough items available to fully fulfill the orders",
            'subsidiary' => "NetSuite order subsidiary internal ID",
            'department' => "NetSuite order department internal ID",
            'opportunity' => "NetSuite order opportunity internal ID",
            'class' => "NetSuite order class internal ID",
            'orderstatus' => "NetSuite order orderstatus internal ID",
            'shippingtaxcode' => "NetSuite shipping tax code internal ID",
        ];
        //This array contains all required conditional mappings
        private static $prefinedCondMapping = [
//        [
//            'type' => 2,
//            'operator' => 1,
//            'wc_field_key' => 'customer_role',
//            'wc_field_value' => 'customer',
//            'ns_field_key' => 'recordType',
//            'ns_field_value' => '',
//            'required' => 1,
//            'depends_on_configuration' => 1
//        ]
        ];

        public function __construct() {
            //tab link and content
            add_filter('in8_admin_tab', [$this, 'addTab']);
            add_action('in8_netsuite_top_level_tabs_page', [$this, 'tabContent']);

            //ajax for saving plugin admin settings
            add_action('wp_ajax_' . self::getOptionsTag(), [$this, 'saveForm']);

            //ajax for saving plugin admin settings
            add_action('wp_ajax_' . self::$optionsTagCO, [$this, 'saveConfigurableOptionForm']);

            //ajax for getting saved forms and payment methods
            add_action('wp_ajax_' . self::$optionsTagFP, [$this, 'getFormsAndPaymentMethodsFromNS']);
        }

        public function addTab($tabs) {
            $tabs[self::$optionsTag] = ['label' => "Sales Order", "priority" => 5];
            return $tabs;
        }

        public function tabContent() {
            global $wpdb;
            $options = In8_SalesOrder::options();

//this handles logic for putting empty mapping keys
            $default_keys = array_keys(self::$salesOrderFieldMapping);
            $default_values = array_values(self::$salesOrderFieldMapping);

            if (empty($options['mapping_key']) || empty($options['mapping_value'])) {
                $options['mapping_key'] = $default_keys;
                $options['mapping_value'] = $default_values;
            } else {
                foreach ($default_keys as $default_key) {
                    if (!in_array($default_key, $options['mapping_key'])) {
                        $options['mapping_key'][] = $default_key;
                        $options['mapping_value'][] = 0;
                    }
                }
            }
//Woo Payment methods
            $payment_method_map_prefix = "paymentmethod_"; //prefix used in payment method mapping
            $woo_payment_methods = In8_Customer::getActivePaymentGateways();

            foreach ($woo_payment_methods as $woo_payment_method_key => $woo_payment_method) {
                $map_pay_method = $payment_method_map_prefix . $woo_payment_method_key;
                if (!in_array($map_pay_method, $options['mapping_key'])) {
                    $options['mapping_key'][] = $map_pay_method;
                    $options['mapping_value'][] = 0;
                }
            }

            $ship_method_map_prefix = "shipmethod_";
            $woo_shipping_methods = self::getShippingMethods();
	        $woo_shipping_methods = apply_filters('in8_so_shipping_methods_default', $woo_shipping_methods);

            foreach ($woo_shipping_methods as $shipping_method) {
                if (isset($shipping_method['services']) && !empty($shipping_method['services'])) {
                    foreach ($shipping_method['services'] as $service) {
                        $map_ship_method = $ship_method_map_prefix . $service['service_id_on_order'];
                        if (!in_array($map_ship_method, $options['mapping_key'])) {
                            $options['mapping_key'][] = $map_ship_method;
                            $options['mapping_value'][] = 0;
                        }
                    }
                } else {
					$shipping_method['service_id_on_order'] = isset($shipping_method['service_id_on_order']) ? $shipping_method['service_id_on_order'] : '';
                    $map_ship_method = $ship_method_map_prefix . $shipping_method['service_id_on_order'];
                    if (!in_array($map_ship_method, $options['mapping_key'])) {
                        $options['mapping_key'][] = $map_ship_method;
                        $options['mapping_value'][] = 0;
                    }
                }
            }

//Configurable options
            $co_options = self::getOptions(2);
            $woo_order_statuses = wc_get_order_statuses();
//order status mapping for order update
            $mapped_order_statuses = [];

            if (isset($co_options['enable_update']) && $co_options['enable_update'] == 'on') {
                foreach ($woo_order_statuses as $status_key => $woo_order_status) {
                    $found_mapped_key = array_search($status_key, $co_options['status_mapping_key']);
                    if ($found_mapped_key !== FALSE) {
                        $mapped_order_statuses[$status_key] = $co_options['status_mapping_value'][$found_mapped_key];
                    } else {
                        $mapped_order_statuses[$status_key] = '';
                    }
                }
            }

            $helptext = self::$salesOrderFieldHelpText;

//conditional mapping options
            $cm_options = self::getOptions(1);
            $required_key_present = [];
            if (!empty($cm_options)) {
                foreach ($cm_options as $key => $mapping) {
                    $index = $key + 1;

                    if (!empty($mapping['operator']) && ($mapping['operator'] == 2 || !empty($mapping['type']))) {
//                        if (!empty(self::$prefinedCondMapping)) {
//                            foreach (self::$prefinedCondMapping as $pre_mapping_key => $pre_mapping) {
//                                if ($mapping['operator'] == $pre_mapping['operator'] && $mapping['type'] == $pre_mapping['type'] && $mapping['ns_field_key'] == $pre_mapping['ns_field_key']) { //check for required mapping
//                                    $required_key_present[] = $pre_mapping_key;
//                                    $cm_options[$key]['required'] = 1;
//                                }
//                            }
//                        }
                        if ($mapping['operator'] == 2) {
                            $mapping['type'] = 0;
                        }
                        $cm_options[$key]['template'] = ConditionalMapping::getCMTemplate($mapping['type'], $mapping['operator'], $index, $mapping);
                    } else {
                        unset($cm_options[$key]);
                    }
                }
            }

//            if (!empty(self::$prefinedCondMapping)) {
//                foreach (self::$prefinedCondMapping as $pre_mapping_key => $pre_mapping) {
//                    $cm_count = count($cm_options);
//                    if (!in_array($pre_mapping_key, $required_key_present)) {  //insert required key if not present
//                        $cm_options[$cm_count] = $pre_mapping;
//                        $cm_options[$cm_count]['template'] = ConditionalMapping::getCMTemplate($pre_mapping['type'], $pre_mapping['operator'], $cm_count, $pre_mapping);
//                    }
//                }
//            }
//NS order forms, payment methods and shipping methods
            $forms_and_paymentsSO = In8_SalesOrder::getOptions(3);

            if (isset($forms_and_paymentsSO['order_forms']) && !empty($forms_and_paymentsSO['order_forms'])) {
                $order_forms = $forms_and_paymentsSO['order_forms'];
            }

            if (isset($forms_and_paymentsSO['payment_methods']) && !empty($forms_and_paymentsSO['payment_methods'])) {
                $payment_methodsSO = $forms_and_paymentsSO['payment_methods'];
            }

            if (isset($forms_and_paymentsSO['ship_methods']) && !empty($forms_and_paymentsSO['ship_methods'])) {
                $ship_methods = $forms_and_paymentsSO['ship_methods'];
            }



            require_once('admin-setting-tab-salesorder.php');
        }

        //static function for returning option tag
        public static function getOptionsTag() {
            return self::$optionsTag;
        }

        //static function for returning conditional mapping option tag
        public static function getOptionsTagCM() {
            return self::$optionsTagCM;
        }

        //static function for returning configurable options
        public static function getOptionsTagCO() {
            return self::$optionsTagCO;
        }

        /**
         * This functions handles data saving for customers Tab
         */
        public function saveForm() {
            $request = $_REQUEST;
            //if mapping key and mapping values are not empty,
            //this code remove mapping key/value(s) if corresponsing key/value(s) is empty  
            //to eliminate empty values
            if (!empty($request['mapping_key']) && !empty($request['mapping_value'])) {

                for ($i = 0; $i < count($request['mapping_key']); $i++) {

                    if (empty($request['mapping_key'][$i])) {
                        unset($request['mapping_value'][$i]);
                        unset($request['mapping_key'][$i]);
                    }

                    if (empty($request['mapping_value'][$i])) {
                        $request['mapping_value'][$i] = "0";
                    }

                    trim($request['mapping_key'][$i]);
                    trim($request['mapping_value'][$i]);
                }
                $request['mapping_key'] = array_values($request['mapping_key']);
                $request['mapping_value'] = array_values($request['mapping_value']);
            }
            if (!empty($request)) {
                self::saveOptions($request);
            }
            die(json_encode([
                'type' => 'success',
            ]));
        }

        /**
         * This functions handles data saving for sales order Tab
         */
        public function saveConfigurableOptionForm() {
            $request = $_REQUEST;
            //if mapping key and mapping values are not empty,
            //this code remove mapping key/value(s) if corresponsing key/value(s) is empty  
            //to eliminate empty values
            if (is_array($request['co'])) {
                $request = $request['co'];

                if (!isset($request['enable_update'])) {
                    unset($request['status_mapping_key']);
                    unset($request['status_mapping_value']);
                }

                //unset values
                if (isset($request['sync_on_creation'])) {
                    unset($request['sync_on_payment_complete']);
                    unset($request['sync_on_hold']);
                } elseif (isset($request['sync_on_hold'])) {
                    unset($request['sync_on_creation']);
                    unset($request['sync_on_payment_complete']);
                } elseif (isset($request['sync_on_payment_complete'])) {
                    unset($request['sync_on_hold']);
                    unset($request['sync_on_creation']);
                }

                if (isset($request['all_purchase_orders'])) {
                    unset($request['identify_order_type']);
                    unset($request['identify_order_type_custom_field']);
                    unset($request['identify_sales_order_custom_field_value']);
                    unset($request['identify_purchase_order_custom_field_value']);
                }

                if (!isset($request['identify_order_type'])) {
                    unset($request['identify_order_type_custom_field']);
                    unset($request['identify_sales_order_custom_field_value']);
                    unset($request['identify_purchase_order_custom_field_value']);
                }

                self::saveOptions($request, 2); //parameter two -> 0 : simple options, 1 : conditional mapping, 2 : configurable options

                die(json_encode([
                    'type' => 'success'
                ]));
            }

            die(json_encode([
                'type' => 'Error',
                'message' => 'Invalid Request'
            ]));
        }

        public function getFormsAndPaymentMethodsTemplate() {
            $forms_and_payments = In8_SalesOrder::getOptions(3);

            $return_string = "";
            if (isset($forms_and_payments['order_forms']) && !empty($forms_and_payments['order_forms'])) {
                $order_forms = $forms_and_payments['order_forms'];
                $return_string .= '<div class="well"><table class="table table-responsive table-bordered">';
                $return_string .= '<caption><h4>NetSuite Order Forms</h4></caption>';
                $return_string .= '<thead class="warning">';
                $return_string .= '<th>Name</th>';
                $return_string .= '<th>Internal ID</th></thead><tbody>';
                foreach ($order_forms as $order_form) {
                    $return_string .= '<tr><td>' . $order_form['text'] . '</td><td>' . $order_form['internalid'] . '</td></tr>';
                }
                $return_string .= '</tbody></table></div>';
            } else {
                $return_string .= '<div class="well">Error loading custom forms from NetSuite. Please check the error logs.</div>';
            }

            $return_string .= '<div class="well">';
            if (isset($forms_and_payments['payment_methods']) && !empty($forms_and_payments['payment_methods'])) {
                $payment_methods = $forms_and_payments['payment_methods'];
                $return_string .= '<div class="well"><table class="table table-responsive table-bordered"><caption><h4>NetSuite Payment Methods</h4></caption><thead class="warning"><th>Name</th><th>Internal ID</th></thead><tbody>';
                foreach ($payment_methods as $payment_method) {
                    $return_string .= '<tr><td>' . $payment_method['name'] . '</td><td>' . $payment_method['internalId'] . '</td></tr>';
                }
                $return_string .= '</tbody></table>';
            } else {
                $return_string .= '<div class="well">Error loading payment methods from NetSuite. Please check the error logs.</div>';
            }
            $woo_payment_methods = In8_Customer::getActivePaymentGateways();
            if (!empty($woo_payment_methods)) {
                $return_string .='<table class="table table-responsive table-bordered">
                        <caption><h4>WooCommerce Payment Methods</h4></caption>
                        <thead class="warning">
                        <th>ID</th>
                        <th>Title</th>
                    </thead>
                    <tbody>';
                $payment_method_map_prefix = "paymentmethod_"; //prefix used in payment method mapping
                foreach ($woo_payment_methods as $payment_method_id => $woo_payment_method) {
                    $return_string .= "<tr>
                                <td>" . $payment_method_map_prefix . $payment_method_id . "</td>
                                <td>" . $woo_payment_method['title'] . "</td>
                            </tr>";
                }

                $return_string .="</tbody></table>";
            }
            $return_string .="</div>";

            $return_string .= '<div class="well">';
            if (isset($forms_and_payments['ship_methods']) && !empty($forms_and_payments['ship_methods'])) {
                $ship_methods = $forms_and_payments['ship_methods'];

                $return_string .= '<table class="table table-responsive table-bordered"><caption><h4>NetSuite Shipping Methods</h4></caption><thead class="warning"><th>Name</th><th>Internal ID</th></thead><tbody>';
                foreach ($ship_methods as $ship_method) {
                    $return_string .= '<tr><td>' . $ship_method['itemid'] . '</td><td>' . $ship_method['internalId'] . '</td></tr>';
                }
                $return_string .= '</tbody></table>';
            }

            $woo_shipping_methods = self::getShippingMethods();
            if (!empty($woo_shipping_methods)) {
                $ship_method_map_prefix = "shipmethod_";
                $return_string .= '<table class="table table-responsive table-bordered">
                                    <caption><h4>WooCommerce Shipping Methods</h4></caption>
                                    <thead class="warning">
                                    <th>Ship Method ID</th>
                                    <th>Title</th>
                                    </thead>
                                    <tbody>';

                foreach ($woo_shipping_methods as $woo_ship_method_id => $woo_ship_method) {
                    if (!isset($woo_ship_method['services'])) {

                        $return_string .= '<tr>
                                                    <td>' . $ship_method_map_prefix . $woo_ship_method['service_id_on_order'] . '</td><td>' . $woo_ship_method['title'] . '</td></tr>';
                    } else {
                        foreach ($woo_ship_method['services'] as $service) {
                            $return_string .= '<tr>
                                    <td>' . $ship_method_map_prefix . $service['service_id_on_order'] . '</td>
                                    <td>' . $service['service_name'] . '</td></tr>';
                        }
                    }
                }

                $return_string .= '</tbody></table>';
            }
            $return_string .="</div>";

            return $return_string;
        }

        public function getFormsAndPaymentMethodsFromNS() {
//GET NS fields
            $payment_methods = $ship_methods = $order_forms = [];

//get a list of payment methods
            $netsuite_connection = new NetsuiteConnection('0', 'paymentmethod');
            $netsuite_connection_resp = $netsuite_connection->doRequest();
            if (isset($netsuite_connection_resp['success']) && $netsuite_connection_resp['success'] == 1 && !empty($netsuite_connection_resp['data'])) {
                $payment_methods = $netsuite_connection_resp['data'];
            }

//get a list of ship methods
            $netsuite_connection = new NetsuiteConnection('0', 'shipitem');
            $netsuite_connection_resp = $netsuite_connection->doRequest();
            if (isset($netsuite_connection_resp['success']) && $netsuite_connection_resp['success'] == 1 && !empty($netsuite_connection_resp['data'])) {
                $ship_methods = $netsuite_connection_resp['data'];
            }

//get a list of customer forms
            $netsuite_connection = new NetsuiteConnection('0', 'salesorderforms');
            $netsuite_connection_resp = $netsuite_connection->doRequest();
//validate response
            if (isset($netsuite_connection_resp['success']) && $netsuite_connection_resp['success'] == 1 && !empty($netsuite_connection_resp['data'])) {
                $order_forms = $netsuite_connection_resp['data'];
            }

            $options = array();
            $options['order_forms'] = $order_forms;
            $options['payment_methods'] = $payment_methods;
            $options['ship_methods'] = $ship_methods;

            self::saveOptions($options, 3);

            $template = $this->getFormsAndPaymentMethodsTemplate();

            die($template);
        }

        /**
         * @param type $options - array of all options
         * @param type $type - 0 : simple options, 1 : conditional mapping, 2 : configuration option
         * @return array of all options
         */
        public static function saveOptions($options = [], $type = 0) {
            if ($type == 2) {
                $option_tag = self::$optionsTagCO;
            } elseif ($type == 1) {
                $option_tag = self::$optionsTagCM;
            } elseif ($type == 3) {
                $option_tag = self::$optionsTagFP;
            } else {
                $option_tag = self::$optionsTag;
            }

            update_option($option_tag, $options);
            return $options;
        }

        /**
         * 
         * @param type $type - 0 : simple options, 1 : conditional mapping, 2 : configuration option
         * @return array of all options
         */
        public static function getOptions($type = 0) {
            if ($type == 2) {
                $option_tag = self::$optionsTagCO;
            } elseif ($type == 1) {
                $option_tag = self::$optionsTagCM;
            } elseif ($type == 3) {
                $option_tag = self::$optionsTagFP;
            } else {
                $option_tag = self::$optionsTag;
            }

            $options = get_option($option_tag, []);
            if (empty($options)) {
                $options = [];
            }

            return $options;
        }
	    public static function getMappings(){
		    $plugin_options = self::getOptions();
		    $plugin_settings_map        = array_change_key_case( array_combine( $plugin_options['mapping_key'], $plugin_options['mapping_value'] ) );
		    return $plugin_settings_map;
	    }

        public static function options() {
            return self::getOptions(self::getOptionsTag());
        }

        private static function getShippingMethods() {


            global $wpdb;
//getting all shipping methods
            $shipping_methods = WC()->shipping->shipping_methods;
            $ship_methods_data = [];
            $all_services = [];
            if( !empty($shipping_methods) ) {
                foreach ( $shipping_methods as $shipping_method ) {
                    if ( $shipping_method->enabled == 'yes' ) {
                        $ship_methods_data[ $shipping_method->id ]['id']    = $shipping_method->id;
                        $ship_methods_data[ $shipping_method->id ]['title'] = $shipping_method->title;
                
                        if ( $shipping_method->id == 'table_rate' ) {
                            $rates = $wpdb->get_results( "SELECT rate_label,rate_id,shipping_method_id FROM {$shipping_method->rates_table} as rates
    WHERE rates.rate_abort     = '0'" );
                    
                            if ( ! empty( $rates ) ) {
                                foreach ( $rates as $rate ) {
                                    $ship_methods_data[ $shipping_method->id ]['services'][ $rate->rate_id ] = [
                                        'service_id'          => $rate->rate_id,
                                        'method_id'           => $rate->shipping_method_id,
                                        'service_name'        => $rate->rate_label,
                                        'service_id_on_order' => $shipping_method->id . "-" . $rate->shipping_method_id . ":" . $rate->rate_id
                                    ];
                                }
                            }
                        } elseif ( $shipping_method->id == 'ups' || $shipping_method->id == 'fedex_wsdl' || $shipping_method->id == 'fedex' || $shipping_method->id == 'wf_shipping_ups' ) {
                            if ( isset( $shipping_method->settings['services'] ) && ! empty( $shipping_method->settings['services'] ) ) {
                                $services = $shipping_method->settings['services'];
                            } elseif ( isset( $shipping_method->settings['custom_methods'] ) && ! empty( $shipping_method->settings['custom_methods'] ) ) {
                                $services = $shipping_method->settings['custom_methods'];
                            } else {
                                continue;
                            }
                    
                            foreach ( $services as $serviceId => $service ) {
                                if ( $service['enabled'] ) {
                                    $service_name = '';
                                    if ( empty( $service['name'] ) ) {
                                        if ( ! isset( $all_services[ $shipping_method->id ] ) ) {
                                            $myClassReflection = new ReflectionClass( get_class( $shipping_method ) );
                                            if ( $myClassReflection->hasProperty( 'services' ) ) {
                                                $secret = $myClassReflection->getProperty( 'services' );
                                                $secret->setAccessible( TRUE );
                                                $all_services[ $shipping_method->id ] = $secret->getValue( $shipping_method );
                                            } else {
                                                $all_services[ $shipping_method->id ] = [];
                                            }
                                        }
                                
                                        if ( isset( $all_services[ $shipping_method->id ][ $serviceId ] ) ) {
                                            $service_name = $all_services[ $shipping_method->id ][ $serviceId ];
                                        }
                                    } else {
                                        $service_name = $service['name'];
                                    }
                                    if ( empty( $service_name ) ) {
                                        $service_name = $ship_methods_data[ $shipping_method->id ]['title'] . "-service-" . $serviceId;
                                    }
                                    $ship_methods_data[ $shipping_method->id ]['services'][ $serviceId ] = [
                                        'service_id'          => $serviceId,
                                        'service_name'        => $service_name,
                                        'service_id_on_order' => $shipping_method->id . ":" . $serviceId
                                    ];
                                }
                            }
                        } elseif ( $shipping_method->id == 'usps' ) {
                            $serviceId                                                           = "flat_rate_box_priority";
                            $ship_methods_data[ $shipping_method->id ]['services'][ $serviceId ] = [
                                'service_id'          => $serviceId,
                                'service_name'        => '',
                                'service_id_on_order' => $shipping_method->id . ":" . $serviceId
                            ];
                            $serviceId                                                           = "flat_rate_box_express";
                            $ship_methods_data[ $shipping_method->id ]['services'][ $serviceId ] = [
                                'service_id'          => $serviceId,
                                'service_name'        => '',
                                'service_id_on_order' => $shipping_method->id . ":" . $serviceId
                            ];
                    
                    
                            if ( isset( $shipping_method->settings['services'] ) && ! empty( $shipping_method->settings['services'] ) ) {
                                $services = $shipping_method->settings['services'];
                            } else {
                                continue;
                            }
                    
                    
                            foreach ( $services as $serviceId => $service ) {
                                $ship_methods_data[ $shipping_method->id ]['services'][ $serviceId ] = [
                                    'service_id'          => $serviceId,
                                    'service_name'        => '',
                                    'service_id_on_order' => $shipping_method->id . ":" . $serviceId
                                ];
                            }
                        } else {
                            $ship_methods_data[ $shipping_method->id ]['service_id_on_order'] = $shipping_method->id;
                        }
                    }
                }
            }
            return $ship_methods_data;
        }
	
	    public static function testFeature($feature, $check = 'on'){
		    $options = self::getOptions( 2 );
		    
		    return ! empty( $options ) && isset( $options[$feature] ) && $options[$feature] == $check;
	    }

	    public static function getOptionValue($option, $isUpdate = false){
		    $plugin_options = self::getOptions();
		    //combine mapping keys and values and then change keys to lowercaSE
		    $plugin_settings_map        = array_change_key_case( array_combine( $plugin_options['mapping_key'], $plugin_options['mapping_value'] ) );
		    $plugin_settings_update_map = [];
		    
		    if( !$isUpdate ) return isset($plugin_settings_map[$option]) ? $plugin_settings_map[$option] : NULL;
		
		
		    if ( ! empty( $plugin_options['mapping_update'] ) ) { //update
			    $plugin_settings_update_map = array_combine( array_intersect_key( $plugin_options['mapping_key'], $plugin_options['mapping_update'] ), array_intersect_key( $plugin_options['mapping_update'], $plugin_options['mapping_key'] ) );
			    $plugin_settings_update_map = array_change_key_case( $plugin_settings_update_map );
		    }
		    
		    if( isset($plugin_settings_update_map[$option]) ) return NULL;
		
		    return isset($plugin_settings_map[$option]) ? $plugin_settings_map[$option] : NULL;
	    }

    }

    new In8_SalesOrder();

    