<?php
	
	/**
	 * This class handles customer and sales order sync
	 * This class uses order action "woocommerce_checkout_order_processed" which is
	 * triggered right before payment method triggers.This class overwrite the
	 * woocommerce payment calls and calls payment method.If payment call is
	 * successful then it transfer sales order to Netsuite.
	 */
	class In8_Customers_Sales_Sync {
		
		private static $instance = NULL;
		
		static public function getInstance() {
			if ( self::$instance === NULL ) {
				self::$instance = new In8_Customers_Sales_Sync();
			}
			
			return self::$instance;
		}
		
		private $process_payment_on_ns, $is_purchase_order = FALSE; //will be set to true if admin selected payment gateways the order payment gateway
		private $user_id, $order_id, $customer_internal_id = 0, $contact_id = 0;
		private $entry_id = NULL;
		private $credit_card_array = [];
		
		//transaction ID map
		public $transactionIds = [
			'cybersource'       => 'cyber_TXId',
			'sagepaydirect'     => '_VPSTxId',
			'authorize_net_aim' => 'trans_id',
		];
		
		public function __construct() {
			$co_options = In8_SalesOrder::getOptions( 2 );
			
			add_action( 'in8sync_async_sendOrderToNetsuite', [ $this, 'sendOrderToNetsuite' ], 10, 4 );
			add_action( 'in8sync_async_syncOnNetsuite', [ $this, 'newOrderCreated' ], 10, 3 );
			
			global $wp_query;
			if ( ( isset( $wp_query ) && ! is_page( wc_get_page_id( 'checkout' ) ) ) && isset( $co_options['enable_update'] ) && $co_options['enable_update'] == 'on' ) {
				//this hook triggers from admin for updating order to NS on status change
				add_action( 'woocommerce_order_status_changed', [ $this, 'syncOnNetsuite' ], 1, 1 );
			} else {
				// The list of hooks that should send the customer and the order
				$sendOrderHooks    = [];
				$sendCustomerHooks = [ 'woocommerce_checkout_order_processed' ];
				
				foreach ( apply_filters( 'in8_so_sendCustomer_hooks', $sendCustomerHooks ) as $hook ) {
					// If the option to sync in background is active, sync order on the thank you page
					if ( isset( $co_options['sync_async'] ) && $co_options['sync_async'] == 'on' ) {
						add_action( $hook, [ $this, 'handleAsyncCustomer' ] );
					} else {
						add_action( $hook, [ $this, 'newOrderCreated' ] );
					}
				}
				
				if ( ! isset( $co_options['sync_on_creation'] ) ) {
					if ( isset( $co_options['sync_on_hold'] ) && $co_options['sync_on_hold'] == 'on' ) {
						$sendOrderHooks = [ 'woocommerce_order_status_on-hold' ];
					} elseif ( isset( $co_options['sync_cancelled'] ) && $co_options['sync_cancelled'] == 'on' ) {
						$sendOrderHooks = [ 'woocommerce_order_status_cancelled' ];
					} elseif ( isset( $co_options['sync_on_payment_complete'] ) && $co_options['sync_on_payment_complete'] == 'on' ) {
						$sendOrderHooks = [ 'woocommerce_payment_complete', 'woocommerce_order_status_completed' ];
					} else {
						$sendOrderHooks = [ 'woocommerce_order_status_processing' ];
					}
				}
				
				foreach ( apply_filters( 'in8_so_sendOrder_hooks', $sendOrderHooks ) as $hook ) {
					// If the option to sync in background is active, sync order on the thank you page
					if ( isset( $co_options['sync_async'] ) && $co_options['sync_async'] == 'on' ) {
						add_action( $hook, [ $this, 'handleAsyncOrder' ] );
					} else {
						add_action( $hook, [ $this, 'sendOrderToNetsuite' ] );
					}
				}
			}
			
			//make sure token is saved before letting customer and order Sync
			add_filter('in8_do_not_sync', function($default, $order_id, $type){
				$order = wc_get_order($order_id);
				if( !$this->isCyberSourceTokenizationOn($order) ) return $default;
				
				$token = apply_filters('in8_so_token', get_post_meta( $order_id, 'cybersource_token', TRUE ), $order);
				if(
					empty($token)
				){
					add_filter('in8_do_not_sync_error', function(){return 'Do not sync - waiting for tokens. E:765';});
					return true;
				}else{// Base plugin
					if( $type == 'order' ) {
						$test = get_post_meta( $order_id, 'customerSyncToken', TRUE );
						if ( empty( $test ) ) {
							update_post_meta( $order_id, 'customerSyncToken', "T" );
							do_action( 'in8sync_async_syncOnNetsuite', $order_id, FALSE, TRUE );
						}
					}
				}
				
				return $default;
			}, 10, 3);
			
			
			$this->handleFailedOrUnsyncedOrders();
		}
		
		// Using this version to be able to keep backwards compatibility
		//add_filter('in8_do_not_sync_error', function(){return 'Stopped by doNotSync.';});
		public function doNotSync( $order_id, $type = 'order' ) {
			
			if ( In8_SalesOrder::testFeature('do_not_sync_ns_orders') ) {
				
				// Do not sync order if ship via is empty
				if( In8_SalesOrder::testFeature('do_not_sync_ns_orders_created_via') ){
					$emptyvia     = get_post_meta( $order_id, "_created_via", TRUE );
					if ( empty($emptyvia) ) {
						return apply_filters( 'in8_do_not_sync', TRUE, $order_id, $type );
					}
				}
				
				// Do not sync order if payment method is empty
				if( In8_SalesOrder::testFeature('do_not_sync_ns_orders_payment_method') ){
					$emptyvia     = get_post_meta( $order_id, "_payment_method", TRUE );
					if ( empty($emptyvia) ) {
						return apply_filters( 'in8_do_not_sync', TRUE, $order_id, $type );
					}
				}
				
				// Do not resync order/customer for X min
				if( In8_SalesOrder::testFeature('do_not_sync_ns_orders_delay_sync') ){
					// Wait for X time before resyncing
					$already_sent_key = "_in8_{$type}_last_sync";
					$already_sent     = get_post_meta( $order_id, $already_sent_key, TRUE );
					$toWaitTime       = apply_filters('in8_do_not_sync_delay_time', 60, $order_id, $type);
					if ( ! empty( $already_sent ) && time() - $already_sent < $toWaitTime ) {
						// Do not sync
						return apply_filters( 'in8_do_not_sync', TRUE, $order_id, $type );
					} else {
						update_post_meta( $order_id, $already_sent_key, time() );
					}
				}
				
			}
			
			return apply_filters( 'in8_do_not_sync', FALSE, $order_id, $type );
		}
		
		public function handleAsyncCustomer( $order_id ) {
			$order = new WC_Order( $order_id );
			
			$customer_plugin_conf_options = In8_Customer::getOptions( 2 );
			$order_pay_method             = trim( $order->get_payment_method() );
			if (
				//$order_pay_method != 'in8_netsuite_payments' &&
				! empty( $customer_plugin_conf_options ) &&
				
				isset( $customer_plugin_conf_options['enable_credit_card'] ) &&
				$customer_plugin_conf_options['enable_credit_card'] == 'on' &&
				
				isset( $customer_plugin_conf_options['selected_gateway'] ) &&
				trim( $customer_plugin_conf_options['selected_gateway'] ) == $order->get_payment_method()
				
//				&& ( ! empty( $_POST[ $order_pay_method . '-card-cvc' ] ) || ! empty( $_POST[ $order_pay_method . '_accountNumber' ] ) || ! empty( $_POST['cardno'] ) || ! empty( $_POST[ 'icustombilling_gate_card_number' ] ) )
			) {
				$this->newOrderCreated( $order_id );
			} else {
				wp_schedule_single_event( time() + 15, 'in8sync_async_syncOnNetsuite', [ $order_id, FALSE, TRUE ] );
			}
		}
		
		public function handleAsyncOrder( $order_id ) {
			if ( ! apply_filters( 'in8_so_use_async', TRUE, $order_id ) ) {
				$this->sendOrderToNetsuite( $order_id );
				
				return;
			}
			wp_schedule_single_event( time() + 30, 'in8sync_async_sendOrderToNetsuite', [ $order_id, 0, FALSE, TRUE ] );
		}
		
		public function syncOnNetsuite( $order_id ) {
			//check if order has been previously synced to NS
			$order_internal_id = get_post_meta( $order_id, 'internalid', TRUE );
			if ( ! empty( $order_internal_id ) ) { // if order is previously synched to NS
				$this->sendOrderToNetsuite( $order_id, $order_internal_id ); //update order
			}
		}
		
		//This function ftech all processing orders,
		//check if they have been transferred to netsuite and
		//if they weren't sent to netsuite send them to netsuite
		public function processRemainingOrders() {
			global $wpdb;
			$orders = $wpdb->get_results( "SELECT ID FROM {$wpdb->posts} as posts
            WHERE posts.post_type     = 'shop_order' 
            AND posts.post_status   = 'wc-processing'" );
			
			foreach ( $orders as $order ) {
				$orderInternalId = get_post_meta( $order->ID, 'internalid', TRUE );
				if ( empty( $orderInternalId ) ) { //order doesn't exists on Netsuite
					$this->customer_internal_id = 0;
					$this->newOrderCreated( $order->ID );
					if ( ! empty( $this->customer_internal_id ) ) {
						$this->sendOrderToNetsuite( $order->ID );
					}
				}
			}
			
			return;
			//redirecting to in8 plugin
			if ( wp_get_referer() ) {
				wp_safe_redirect( wp_get_referer() );
			} else {
				wp_safe_redirect( get_home_url() );
			}
		}
		
		private function isPurchaseOrder( $order_id ) {
			
			$active_plugins = apply_filters( 'active_plugins', get_option( 'active_plugins' ) );
			
			if ( in_array( 'in8sync-purchaseorder/in8sync-purchaseorder.php', $active_plugins ) ) {
				$plugin_conf_options = In8_PurchaseOrder::getOptions( 2 );
				
				if ( isset( $plugin_conf_options['all_purchase_orders'] ) ) {
					$this->is_purchase_order = TRUE;
				}
				
				if ( $this->is_purchase_order === FALSE && isset( $plugin_conf_options['identify_order_type'] ) && $plugin_conf_options['identify_order_type'] == 'on' ) {
					$order_type = get_post_meta( $order_id, $plugin_conf_options['identify_order_type_custom_field'], TRUE );
					
					if ( $order_type == $plugin_conf_options['identify_purchase_order_custom_field_value'] ) {
						$this->is_purchase_order = TRUE;
					}
				}
			}
			
			if ( $this->is_purchase_order === TRUE ) {
				//remove sales order hooks
				$this->_removeOrderHooks();
			}
			
			return $this->is_purchase_order;
		}
		
		// This function is triggered by woocommerce_checkout_order_processed hook,
		// which triggers when order data is validated and payment method is being called
		public function newOrderCreated( $order_id, $is_sale = FALSE, $reset = FALSE ) {
			$this->shouldReset( $reset );
			
			if ( $this->doNotSync( $order_id, 'customer' ) ) {
				$doNotSyncErrorMessage = apply_filters( 'in8_do_not_sync_error', FALSE, $order_id, 'customer' );
				if( !empty($doNotSyncErrorMessage) ){
					$netsuite_connection = new NetsuiteConnection( $order_id, 'salesorder-customer' );
					$netsuite_connection->processError( $doNotSyncErrorMessage );
				}
				return;
			}
			
			// Allows to extend and set a specific customer id and bypass the customer sync
			if ( ( $cId = apply_filters( 'in8sync_so_order_sync_customer', 0, $order_id ) ) != 0 ) {
				$this->customer_internal_id = $cId;
				
				// Updates the customerinternalid field according to the overriden Id
				update_post_meta( $order_id, "customerinternalid", $this->customer_internal_id );
				
				return FALSE;
			}
			
			if ( ! $is_sale ) {
				if ( $this->isPurchaseOrder( $order_id ) ) {
					return;
				}
			}
			
			$customer_plugin_conf_options = In8_Customer::getOptions( 2 );
			$this->order_id               = $order_id;
			
			$order            = new WC_Order( $this->order_id );
			$order_pay_method = $order->get_payment_method();
			
			/**
			 * This will halt this hook in case this payment method need to sync after payment
			 */
			if ( is_page( wc_get_page_id( 'checkout' ) ) && current_filter() == "woocommerce_checkout_order_processed" && ! empty( $customer_plugin_conf_options['payment_method_check'] ) && is_array( $customer_plugin_conf_options['payment_method_check'] ) &&
			     in_array( $order_pay_method, $customer_plugin_conf_options['payment_method_check'] ) ) {
				return FALSE;
			}
			
			
			if ( ! class_exists( 'NetsuiteConnection' ) ) { //in case of ajax
				//netsuite api connection file
				require_once( WP_PLUGIN_DIR . '/in8sync/code/netsuiteconnection.php' );
			}
			
			$netsuite_connection = new NetsuiteConnection( $this->order_id, 'salesorder-customer' );
			
			
			//set user id to set zero in case admin is creating the order
			$this->user_id = $order->get_customer_id();
			//customer email
			$customer_email = $order->get_billing_email();
			//plugin options
			$plugin_options = In8_Customer::getOptions();
			
			
			$billing_company_field = $order->get_billing_company();
			$contactSearchIsOn     = ! empty( $billing_company_field ) && isset( $customer_plugin_conf_options['enable_contact_sync'] ) && $customer_plugin_conf_options['enable_contact_sync'] == 'on';
			$include_company       = isset( $customer_plugin_conf_options['search_include_company'] ) ? $customer_plugin_conf_options['search_include_company'] : FALSE;
			$include_phone         = isset( $customer_plugin_conf_options['search_include_phone'] ) ? $customer_plugin_conf_options['search_include_phone'] : FALSE;
			
			// Check if the customer internal id is saved on the order
			if ( empty( $this->customer_internal_id ) ) {
				$this->customer_internal_id = get_post_meta( $this->order_id, 'customerinternalid', TRUE );
				$this->contact_id           = get_post_meta( $this->order_id, 'contactinternalid', TRUE );
				if ( ! empty( $this->customer_internal_id ) ) {
					
					// UPDATE NETSUITE CUSTOMER DATA
					$customer_data = $this->_prepareCustomerData( $order, $plugin_options );
					
					if ( $customer_data !== FALSE ) {
						unset( $customer_data['datecreated'] );
						$customer_data['internalId'] = $this->customer_internal_id;
						
						// update contact if needed
						if ( $contactSearchIsOn && ! empty( $this->contact_id ) ) {
							$this->sync_contact( $customer_data );
							// restrict company data
							$customer_data = $this->_getCompanyData( $customer_data );
						}
						// update netsuite customer/company
						$customer_response = $netsuite_connection->doRequest( $customer_data );
						
						if ( $customer_response && isset( $customer_response['data']['id'] ) ) {
							
							// When updating the customer data, if the RESTlet returns a different customer id, needs to use the returned Id and update the meta field
							// Cases which the customer is in a different subsidiary or if the customer gets deleted from NetSuite
							if ( $this->customer_internal_id != $customer_response['data']['id'] ) {
								$this->customer_internal_id = $customer_response['data']['id'];
								update_post_meta( $this->order_id, "customerinternalid", $this->customer_internal_id );
							}
						}
						
						$auth_successful = get_post_meta( $this->order_id, "NS_auth_successful", TRUE );
						if ( $auth_successful = 'no' ) {
							if ( ! empty( $customer_response['creditCards'] ) ) {
								foreach ( $customer_response['creditCards'] as $cc_response ) {
									if ( $cc_response['ccdefault'] == 'T' ) {
										if ( ! empty( $cc_response['ccinternalid'] ) ) {
											update_post_meta( $order->get_id(), 'ns_ccinternalid', $cc_response['ccinternalid'] );
										}
									}
								}
							}
						}
					}
				}
			}
			
			// Check if the customerinternalid is on the user, if it's not a guest order
			if ( empty( $this->customer_internal_id ) && $this->user_id !== 0 ) {
				// get from user data
				$this->customer_internal_id = get_user_meta( $this->user_id, 'customerinternalid', TRUE );
				$this->contact_id           = get_user_meta( $this->user_id, 'contactinternalid', TRUE );
				
				if ( ! empty( $this->customer_internal_id ) ) {
					
					// UPDATE NETSUITE CUSTOMER DATA
					$customer_data = $this->_prepareCustomerData( $order, $plugin_options );
					if ( $customer_data !== FALSE ) {
						unset( $customer_data['datecreated'] );
						
						$customer_data['internalId'] = $this->customer_internal_id;
						// update contact if needed
						if ( $contactSearchIsOn && ! empty( $this->contact_id ) ) {
							$this->sync_contact( $customer_data );
							// restrict company data
							$customer_data = $this->_getCompanyData( $customer_data );
						}
						// update netsuite customer/company
						$customer_response = $netsuite_connection->doRequest( $customer_data );
						
						// When updating the customer data, if the RESTlet returns a different customer id, needs to use the returned Id
						// Update the user meta field
						if ( $customer_response && isset( $customer_response['data']['id'] ) ) {
							if ( $this->customer_internal_id != $customer_response['data']['id'] ) {
								$this->customer_internal_id = $customer_response['data']['id'];
								update_user_meta( $this->user_id, 'customerinternalid', $this->customer_internal_id );
							}
						}
					}
				}//{"success":true,"data":[{"internalId":"63632","altcontact":"","altemail":"","altphone":"","attention":"Robert Staging","balance":0,"category":"","comments":"","companyname":"In8Sync Staging","credithold":{"name":"Auto","value":"AUTO"},"creditholdoverride":false,"creditlimit":"","currency":{"name":"USA","internalId":1},"dateclosed":"12/15/2017 10:29 am","datecreated":"12/15/2017 10:29 am","daysoverdue":0,"defaultorderpriority":"","depositbalance":0,"email":"stagingcontact@in8sync.com","emailpreference":{"name":"Default","value":"DEFAULT"},"emailtransactions":false,"enddate":"","entityid":"110","entitynumber":"110","entitystatus":{"name":"CUSTOMER-Closed Won","internalId":13},"externalid":"","fax":"","firstname":"","firstsaledate":"","giveaccess":false,"homephone":"","image":"","isinactive":false,"isperson":false,"lastsaledate":"","middlename":"","mobilephone":"","overduebalance":0,"parent":{"name":"110 In8Sync Staging","internalId":63632},"phone":"","salesrep":{"name":"Robert Lasater","internalId":-5},"shippingcarrier":{"name":"FedEx/USPS/More","value":"nonups"},"shippingitem":"","stage":{"name":"Customer","value":"CUSTOMER"},"startdate":"","taxitem":"","terms":{"name":"Net 30","internalId":2},"addresses":[]},{"internalId":"63633","firstname":"","lastname":"","middlename":"","phone":"","email":"stagingcontact@in8sync.com","company":{"name":"110 In8Sync Staging","internalId":63632},"isContact":true}],"errors":[]}
			}
			
			// SEARCH 1 - search for contact if contacts are enabled or search for customer if contact is not enabled, by email and company name(if company name is enabled)
			if (empty($this->customer_internal_id) || ( $contactSearchIsOn && empty($this->contact_id) ) ) {
				
				$search_by = 'customer';
				if ( $contactSearchIsOn ) {
					$search_by = 'contact';
				}
				$netsuite_search_connection = new NetsuiteConnection( $this->order_id, 'salesorder-' . $search_by . '-search' );
				
				//get filter expression for search
				$filterExpression = CustomerHelperFunctions::getCustomerFilterExpression( $customer_email, $include_company, $order->billing_company, $include_phone, $order->billing_phone, $customer_plugin_conf_options );
				//customer search request
				$customer_search = $netsuite_search_connection->doRequest( [], [ 'filterExpression' => $filterExpression ] );
				//if search returns result
				if ( $customer_search['success'] == 1 && ! empty( $customer_search['data'] ) ) {
					do_action( 'in8_search_entities_results', $customer_search['data'], $customer_email, $order_id );
					// if it returns 1 or more results we need to make sure they follow the other filters(email, company, phone) if any of them are enabled
					if ( count( $customer_search['data'] ) >= 1 ) {
						//filter search to get internal id
						if ( $contactSearchIsOn ) {
							$item_result = CustomerHelperFunctions::processContactSearchResult( $customer_search['data'], $customer_email, $include_company, $order->billing_company, $include_phone, $order->billing_phone, FALSE );
						} else {
							$item_result = CustomerHelperFunctions::processCustomerSearchResult( $customer_search['data'], $customer_email, $customer_plugin_conf_options['search_include_company'], $order->billing_company, $customer_plugin_conf_options['search_include_phone'], $order->billing_phone, FALSE );
						}
						// Checks if the return is a contact information
						if ( $item_result['isContact'] == TRUE ) {
							$this->customer_internal_id = $item_result['company']['internalId'];
							$this->contact_id           = $item_result['internalId'];
						} else {
							$this->customer_internal_id = $item_result['internalId'];  //customer update on email search
						}
					}
					
					//customer data array
					if ( ! empty( $this->customer_internal_id ) ) {
						$customer_data = $this->_prepareCustomerData( $order, $plugin_options );
						
						if ( $customer_data !== FALSE ) {
							
							unset( $customer_data['datecreated'] );
							$customer_data['internalId'] = $this->customer_internal_id;
							
							// update contact if needed
							$customer_data['wc_search_1'] = 1;
							if ( $contactSearchIsOn && ! empty( $this->contact_id ) ) {
								$this->sync_contact( $customer_data );
								// restrict company data
								$customer_data = $this->_getCompanyData( $customer_data );
							}
							// update netsuite customer/company
							$customer_response = $netsuite_connection->doRequest( $customer_data );
							
							// When updating the customer data, if the RESTlet returns a different customer id (different subsidiary), needs to use the returned Id
							if ( $customer_response && isset( $customer_response['data']['id'] ) && ! $this->contact_id ) {
								$this->customer_internal_id = $customer_response['data']['id'];
							}
						}
					}
				}
				
				
				// contact was not found, search for customer / only needed if contact search is on
				if ( empty( $this->customer_internal_id ) && $contactSearchIsOn ) {
					
					$search_by                  = 'customer';
					$netsuite_search_connection = new NetsuiteConnection( $this->order_id, 'salesorder-' . $search_by . '-search' );
					
					//get filter expression for search
					$filterExpression = CustomerHelperFunctions::getCustomerFilterExpression( $customer_email, $include_company, $order->billing_company, $include_phone, $order->billing_phone );
					//customer search request
					$customer_search = $netsuite_search_connection->doRequest( [], [ 'filterExpression' => $filterExpression ] );
					//if serach returns result
					if ( $customer_search['success'] == 1 && ! empty( $customer_search['data'] ) ) {
						do_action( 'in8_search_entities_results', $customer_search['data'], $customer_email, $order_id );
						// if it returns 1 or more results we need to make sure they follow the other filters(email, company, phone) if any of them are enabled
						if ( count( $customer_search['data'] ) >= 1 ) {
							//filter search to get internal id
							$item_result = CustomerHelperFunctions::processCustomerSearchResult( $customer_search['data'], $customer_email, $include_company, $order->billing_company, $include_phone, $order->billing_phone, FALSE );
							// Checks if the return is a contact information
							if ( $item_result['isContact'] == TRUE ) {
								$this->customer_internal_id = $item_result['company']['internalId'];
								$this->contact_id           = $item_result['internalId'];
							} else {
								$this->customer_internal_id = $item_result['internalId'];  //customer update on email search
							}
						}
//						elseif ( count( $customer_search['data'] ) == 1 ) {
//
//							// Checks if the return is a contact information
//							if ( $customer_search['data'][0]['isContact'] == TRUE ) {
//								$this->customer_internal_id = $customer_search['data'][0]['company']['internalId'];
//								$this->contact_id           = $customer_search['data'][0]['internalId'];
//							} else {
//								$this->customer_internal_id = $customer_search['data'][0]['internalId'];  //customer update on email search
//							}
//						}
						
						// if customer was found
						if ( ! empty( $this->customer_internal_id ) ) {
							// sync contact if contact sync is on
							$customer_data                = $this->_prepareCustomerData( $order, $plugin_options );
							$customer_data['wc_search_2'] = 1;
							if ( $contactSearchIsOn ) {
								if ( ! $this->sync_contact( $customer_data ) ) {
									$error = "ERROR :: Contact failed creating.";
									$netsuite_connection->processError( $error );
									
									return;
								}
								$customer_data = $this->_getCompanyData( $customer_data );
							}
							
							// update existing customer
							$customer_response = $netsuite_connection->doRequest( $customer_data );
							
							// When updating the customer data, if the RESTlet returns a different customer id (different subsidiary), needs to use the returned Id
							if ( $customer_response && isset( $customer_response['data']['id'] ) && ! $this->contact_id ) {
								$this->customer_internal_id = $customer_response['data']['id'];
							}
						}
					}
				}
				
				// Search by domain name
				if ( empty( $this->customer_internal_id ) && $contactSearchIsOn && isset( $customer_plugin_conf_options['search_email_domain'] ) && $customer_plugin_conf_options['search_email_domain'] == 'on' ) {
					$netsuite_search_connection = new NetsuiteConnection( $this->order_id, 'salesorder-customer-search' );
					
					//get filter expression for search
					$filterExpression = CustomerHelperFunctions::getEmailEndingFilterExpression( $customer_email, $customer_plugin_conf_options );
					
					//customer search request
					$customer_search = $netsuite_search_connection->doRequest( [], [ 'filterExpression' => $filterExpression ] );
					if ( $customer_search['success'] == 1 && ! empty( $customer_search['data'] ) ) {
						//if email search returns more than one result
						if ( count( $customer_search['data'] ) > 1 ) {
							//filter search to get internal id
							$this->customer_internal_id = CustomerHelperFunctions::processCustomerSearchResult( $customer_search['data'], $customer_email, $include_company, $order->billing_company, $include_phone, $order->billing_phone );
							if ( empty( $this->customer_internal_id ) ) {
								$this->customer_internal_id = $customer_search['data'][0]['internalId'];
							}
						} elseif ( count( $customer_search['data'] ) == 1 ) {
							$this->customer_internal_id = $customer_search['data'][0]['internalId'];
						}
						
						// create the new contact
						$customer_data                = $this->_prepareCustomerData( $order, $plugin_options );
						$customer_data['wc_search_3'] = 1;
						if ( ! $this->sync_contact( $customer_data ) ) {
							$error = "ERROR :: Contact failed creating.";
							$netsuite_connection->processError( $error );
							
							return;
						}
					}
				}
				
				// customer was not found, create a new one
				if ( empty( $this->customer_internal_id ) ) {
					$customer_data = $this->_prepareCustomerData( $order, $plugin_options );
					if ( $customer_data !== FALSE ) {
						unset( $customer_data['datecreated'] );
						
						$customer_data['wc_search_4'] = 1;
						$customer_response            = $netsuite_connection->doRequest( $customer_data );
						// When updating the customer data, if the RESTlet returns a different customer id, needs to use the returned Id
						// Update the user meta field
						if ( $customer_response && isset( $customer_response['data']['id'] ) ) {
							if ( $this->customer_internal_id != $customer_response['data']['id'] ) {
								$this->customer_internal_id = $customer_response['data']['id'];
								
								// create the contact for the customer, if settings are on
								if ( $contactSearchIsOn ) {
									$customer_data = $this->_prepareCustomerData( $order, $plugin_options );
									if ( ! $this->sync_contact( $customer_data ) ) {
										$error = "ERROR :: Contact failed creating.";
										$netsuite_connection->processError( $error );
										
										return;
									}
								}
							}
						}
					}
				}
				
				$customer_id = $order->get_customer_id();
				if ( ! empty( $this->customer_internal_id ) && empty( $customer_id ) ) {
					update_post_meta( $order->get_id(), 'customerinternalid', $this->customer_internal_id );
					if ( ! empty( $this->contact_id ) ) {
						update_post_meta( $order->get_id(), 'contactinternalid', $this->contact_id );
					}
				}
			}
			
			// if we only found the customer and the contact sync is on we need to create a new contact
			if ( ! empty( $this->customer_internal_id ) && empty( $this->contact_id ) && $contactSearchIsOn ) {
				$customer_data                = $this->_prepareCustomerData( $order, $plugin_options );
				$customer_data['wc_search_5'] = 1;
				if ( ! $this->sync_contact( $customer_data ) ) {
					$error = "ERROR :: Contact failed creating.";
					$netsuite_connection->processError( $error );
					
					return;
				}
			}
			
			
			//save customer internal id in user meta fields
			if ( ! empty( $this->customer_internal_id ) ) {
				//save user meta only if user is logged in
				if ( $this->user_id !== 0 ) {
					update_user_meta( $this->user_id, "customerinternalid", $this->customer_internal_id, TRUE );
					
					if ( $this->contact_id ) {
						update_user_meta( $this->user_id, "contactinternalid", $this->contact_id, TRUE );
					}
				}
				//save order meta data
				add_post_meta( $this->order_id, "customerinternalid", $this->customer_internal_id, TRUE );
				if ( $this->contact_id ) {
					update_post_meta( $this->order_id, "contactinternalid", $this->contact_id, TRUE );
				}
				
				$order_co_options = In8_SalesOrder::getOptions( 2 );
				
				if ( ! empty( $_REQUEST['reprocess_remaining_orders'] ) ) {
					return FALSE;
				}
				//if on checkout page and Pyament has to be made on NS
				if ( $this->process_payment_on_ns || ( isset( $order_co_options['sync_on_creation'] ) && $order_co_options['sync_on_creation'] == 'on' ) ) {
					//send order to NetSuite
					$this->sendOrderToNetsuite( $this->order_id, 0, TRUE );
				}
			} else {
				
				if ( is_admin() ) {
					return FALSE;
				}
				
				//check if order's payment method is same as admin defined method for this plugin
				if ( ! empty( $customer_plugin_conf_options ) &&
				     isset( $customer_plugin_conf_options['enable_credit_card'] ) &&
				     $customer_plugin_conf_options['enable_credit_card'] == 'on' &&
				     isset( $customer_plugin_conf_options['selected_gateway'] ) &&
				     trim( $customer_plugin_conf_options['selected_gateway'] ) == trim( $order_pay_method ) ) {
					//get order data and order payment method
					// If customer with payment method of Credit Credit fails to make it into NetSuite, force message to try again or another payment method.
					$error = "ERROR :: Customer can't be created in NetSuite so forcing to select another payment gateway";
					$netsuite_connection->processError( $error );
					
					
					echo json_encode( [
						'result'   => 'failure',
						"messages" => "<ul class='woocommerce-error'><li>This payment method is not working, please try another</li></ul>",
						'refresh'  => TRUE,
						'reload'   => FALSE,
					] );
					die;
				} else {
					$error = "Order can't be sent to NetSuite as Customer can't be created";
					$netsuite_connection->processError( $error );
					//remove further action as order can't be synced to NS withput customer
					$this->_removeOrderHooks();
				}
			}
		}
		
		private function _removeOrderHooks() {
			remove_action( 'woocommerce_order_status_on-hold', [ $this, 'sendOrderToNetsuite' ], 1 );
			remove_action( 'woocommerce_order_status_processing', [ $this, 'sendOrderToNetsuite' ], 1 );
			remove_action( 'woocommerce_payment_complete', [ $this, 'sendOrderToNetsuite' ], 1 );
		}
		
		public function sync_contact( $contact_data ) {
			$co_options = In8_Customer::getOptions( 2 );
			if ( ! isset( $co_options['enable_contact_sync'] ) || $co_options['enable_contact_sync'] != 'on' ) {
				return FALSE;
			}
			
			if ( empty( $this->customer_internal_id ) ) {
				return FALSE;
			}
			
			$netsuite_connection = new NetsuiteConnection( $this->customer_internal_id, 'customer-contact' );
			
			unset( $contact_data['category'] );
			unset( $contact_data['customform'] );
			$contact_data['recordType'] = 'contact';
			$contact_data['company']    = $this->customer_internal_id;
			unset( $contact_data['internalId'] );
			if ( ! empty( $this->contact_id ) ) {
				$contact_data['internalId'] = $this->contact_id;
			}
			
			$contact_data['isperson'] = 'T';
			
			$contact_data = apply_filters( 'in8sync_before_contact_sync', $contact_data, $this->customer_internal_id, $this->user_id, $this->order_id );
			
			if ( $contact_data !== FALSE ) {
				$customer_response = $netsuite_connection->doRequest( $contact_data );
				if ( $customer_response && isset( $customer_response['data']['id'] ) ) {
					$this->contact_id = $customer_response['data']['id'];
					do_action( 'in8sync_after_contact_sync', $this->contact_id, $customer_response, $this->user_id );
					
					return TRUE;
				}
			}
			
			return FALSE;
		}
		
		/**
		 * Allows us to restrict what data can be synced with the company
		 *
		 * @param $customer_data
		 *
		 * @return array
		 */
		public function _getCompanyData( $customer_data ) {
			if ( empty( $this->contact_id ) ) {
				return $customer_data;
			}
			
			$remove_fields = [
				'email',
				'firstname',
				'lastname',
				'phone',
				'companyname',
				'isperson',
			];
			
			if ( ! empty( $this->customer_internal_id ) ) {
				$remove_fields[] = 'custentity_in8_wc_customer_id';
			}
			
			$remove_fields = apply_filters( 'in8_so_company_remove_fields', $remove_fields );
			
			foreach ( $remove_fields as $field ) {
				if ( isset( $customer_data[ $field ] ) ) {
					// moving the data to company_data in case we need them latter
					if( !isset($customer_data['company_data']) ){
						$customer_data['company_data'] = [];
					}
					$customer_data['company_data'][$field] = $customer_data[ $field ];
					
					unset( $customer_data[ $field ] );
				}
			}
			
			return apply_filters( 'in8_so_update_company', $customer_data );
		}
		
		/**
		 * This function prepare customer data for adding/updating on netsuite
		 */
		public function _prepareCustomerData( $order, $plugin_options ) {
			$plugin_settings_map = [];
			
			if ( ! empty( $plugin_options['mapping_key'] ) && ! empty( $plugin_options['mapping_value'] ) ) {
				$plugin_settings_map = array_combine( $plugin_options['mapping_key'], $plugin_options['mapping_value'] );
				$plugin_settings_map = array_change_key_case( $plugin_settings_map );
				if ( ! empty( $this->customer_internal_id ) && ! empty( $plugin_options['mapping_update'] ) ) {
					$plugin_settings_update_map = array_combine( array_intersect_key( $plugin_options['mapping_key'], $plugin_options['mapping_update'] ), array_intersect_key( $plugin_options['mapping_update'], $plugin_options['mapping_key'] ) );
					$plugin_settings_update_map = array_change_key_case( $plugin_settings_update_map );
				}
			}
			//plugin conf options
			
			$customer_plugin_conf_options = In8_Customer::getOptions( 2 );
			
			$currency = NULL;
			
			// Get WooCommerce default currency
			$order_currency = strtolower( get_woocommerce_currency() );
			
			// Get the currency applied to the order
			$on_order_currency = get_post_meta( $this->order_id, '_order_currency', TRUE );
			
			// If the currency applied to order is not empty, use it
			$order_currency = ! empty( $on_order_currency ) ? strtolower( $on_order_currency ) : $order_currency;
			
			if ( empty( $this->customer_internal_id ) || ( ! empty( $this->customer_internal_id ) && ! isset( $plugin_settings_update_map[ 'currency_' . $order_currency ] ) ) ) {
				if ( isset( $customer_plugin_conf_options['currency'] ) && $customer_plugin_conf_options['currency'] == 'on' ) {
					//order currency
					if ( ! empty( $plugin_settings_map[ 'currency_' . $order_currency ] ) ) {
						//currency array
						$currency = [ "currency" => $plugin_settings_map[ 'currency_' . $order_currency ] ];
					}
				}
			}
			
			$order_pay_method = $order->get_payment_method();
			$order_pay_method_alias = $order_pay_method == 'cybersource_credit_card' ? 'wc-cybersource-credit-card' : $order_pay_method;
			if( $order_pay_method == 'cybersource' ){
				$order_pay_method_alias = 'wc-cybersource';
			}
			
			//custome & order email
			$customer_email = $order->get_billing_email();
			//extract billing address data
			$first_name = $order->get_billing_first_name();
			$last_name  = $order->get_billing_last_name();
			$company    = $order->get_billing_company();
			
			$address_1 = $order->get_billing_address_1();
			$address_2 = $order->get_billing_address_2();
			$city      = $order->get_billing_city();
			$state     = $order->get_billing_state();
			$country   = $order->get_billing_country();
			$postcode  = $order->get_billing_postcode();
			$phone     = $order->get_billing_phone();
			$bill_company = $order->get_billing_company();
			
			//IS PERSON
			if ( ! empty( $company ) || isset( $customer_plugin_conf_options['disable_is_person'] ) && $customer_plugin_conf_options['disable_is_person'] == 'on' ) {
				$is_person = "F";
				if ( empty( $company ) ) {
					$company = $first_name . ' ' . $last_name;
				}
			} else {
				$is_person = "T";
			}
			//custom form
			$customform = isset( $plugin_settings_map['customform'] ) ? $plugin_settings_map['customform'] : '';
			
			//check if order's payment method is same as admin defined method for this plugin
			if ( ! empty( $customer_plugin_conf_options ) &&
			     isset( $customer_plugin_conf_options['enable_credit_card'] ) &&
			     $customer_plugin_conf_options['enable_credit_card'] == 'on'
				//                    isset($customer_plugin_conf_options['selected_gateway'])
				//&& trim($plugin_options['selected_gateway']) == trim($order_pay_method)
			) {
				$ccData = apply_filters('in8sync_so_cc_data', [], $order);
				
				//credit card info for wordlpay
				if ( ! empty( $_POST['expm'] ) &&
				     ! empty( $_POST['expy'] ) &&
				     ! empty( $_POST['cardno'] ) && ! empty( $_POST['cvv'] ) ) {
					
					$originalDate = $_POST['expy'] . "-" . $_POST['expm'] . "-01";
					$originalDate = date( "m/Y", strtotime( $originalDate ) );
					$card_type    = $this->_cardType( $_POST['cardno'] );
					$pay_method   = 'paymentmethod_' . strtolower( $_POST['paymentmethod'] );
					$card_name    = ! empty( $_POST['card_holder'] ) ? $_POST['card_holder'] : $first_name . " " . $last_name;
					if ( $card_type !== 0 && ! empty( $plugin_settings_map[ $pay_method ] ) ) {
						
						//credit card array
						$this->credit_card_array = [
							'ccnumber'      => $_POST['cardno'],
							'paymentmethod' => $plugin_settings_map[ $pay_method ],
							'ccexpiredate'  => $originalDate,
							'ccname'        => $card_name,
							'ccdefault'     => "T",
						];
					}
				} elseif ( ! empty( $_POST[ $order_pay_method . '_accountNumber' ] ) &&
				           ! empty( $_POST[ $order_pay_method . '_expirationMonth' ] ) &&
				           ! empty( $_POST[ $order_pay_method . '_expirationYear' ] ) &&
				           ! empty( $_POST[ $order_pay_method . '_cvNumber' ] ) ) {
					
					$originalDate = $_POST[ $order_pay_method . '_expirationYear' ] . "-" . $_POST[ $order_pay_method . '_expirationMonth' ] . "-01";
					
					$originalDate = date( "m/Y", strtotime( $originalDate ) );
					
					$card_type = $this->_cardType( $_POST[ $order_pay_method . '_accountNumber' ] );
					
					$pay_method = 'paymentmethod_' . $card_type;
					
					if ( $card_type !== 0 && ! empty( $plugin_settings_map[ $pay_method ] ) ) {
						//credit card array
						$this->credit_card_array = [
							'ccnumber'      => $_POST[ $order_pay_method . '_accountNumber' ],
							'paymentmethod' => $plugin_settings_map[ $pay_method ],
							'ccexpiredate'  => $originalDate,
							'ccname'        => $first_name . " " . $last_name,
							'ccdefault'     => "T",
						];
					}
				} elseif ( ! empty( $_POST[ $order_pay_method . '-card-cvc' ] ) &&
				           ! empty( $_POST[ $order_pay_method . '-card-expiry' ] ) &&
				           ! empty( $_POST[ $order_pay_method . '-card-number' ] ) ) {
					
					list( $month, $year ) = explode( "/", $_POST[ $order_pay_method . '-card-expiry' ] );
					$month = trim( $month );
					$year  = trim( $year );
					
					$originalDate = $year . "-" . $month . "-01";
					
					$originalDate = date( "m/Y", strtotime( $originalDate ) );
					
					$card_type = $this->_cardType( $_POST[ $order_pay_method . '-card-number' ] );
					
					$pay_method = 'paymentmethod_' . $card_type;
					
					if ( $card_type !== 0 && ! empty( $plugin_settings_map[ $pay_method ] ) ) {
						//credit card array
						$this->credit_card_array = [
							'ccnumber'      => $_POST[ $order_pay_method . '-card-number' ],
							'paymentmethod' => $plugin_settings_map[ $pay_method ],
							'ccexpiredate'  => $originalDate,
							//'ccmemo'        => $_POST[ $order_pay_method . '-card-cvc' ],
							'ccname'        => $first_name . " " . $last_name,
							'ccdefault'     => "T",
						];
					}
				} elseif ( ! empty( $_POST[ $order_pay_method_alias . '-csc' ] ) &&
				           ! empty( $_POST[ $order_pay_method_alias . '-expiry' ] ) &&
				           ! empty( $_POST[ $order_pay_method_alias . '-account-number' ] ) ) {
					
					list( $month, $year ) = explode( "/", $_POST[ $order_pay_method_alias . '-expiry' ] );
					$month = trim( $month );
					$year  = trim( $year );
					
					$originalDate = $year . "-" . $month . "-01";
					
					$originalDate = date( "m/Y", strtotime( $originalDate ) );
					
					$card_type = $this->_cardType( $_POST[ $order_pay_method_alias . '-account-number' ] );
					
					$pay_method = 'paymentmethod_' . $card_type;
					
					if ( $card_type !== 0 && ! empty( $plugin_settings_map[ $pay_method ] ) ) {
						//credit card array
						$this->credit_card_array = [
							'ccnumber'      => $_POST[ $order_pay_method_alias . '-account-number' ],
							'paymentmethod' => $plugin_settings_map[ $pay_method ],
							'ccexpiredate'  => $originalDate,
							//'ccmemo'        => $_POST[ $order_pay_method . '-card-cvc' ],
							'ccname'        => $first_name . " " . $last_name,
							'ccdefault'     => "T",
						];
					}
				} elseif ( ! empty( $_POST[ 'icustombilling_gate_card_csc' ] ) &&
				           ! empty( $_POST[ 'icustombilling_gate_card_expiration' ] ) &&
				           ! empty( $_POST[ 'icustombilling_gate_card_number' ] ) ) {
					list( $month, $year ) = explode( "/", $_POST[ 'icustombilling_gate_card_expiration' ] );
					$month = trim( $month );
					$year  = trim( $year );
					
					$originalDate = $year . "-" . $month . "-01";
					
					$originalDate = date( "m/Y", strtotime( $originalDate ) );
					
					$cardNumber    = $_POST[ 'icustombilling_gate_card_number' ];
					$cardNumber = str_replace('-', '', $cardNumber);
					$card_type = $this->_cardType( $cardNumber );
					
					$pay_method = 'paymentmethod_' . $card_type;
					
					if ( $card_type !== 0 && ! empty( $plugin_settings_map[ $pay_method ] ) ) {
						//credit card array
						$this->credit_card_array = [
							'ccnumber'      => $cardNumber,
							'paymentmethod' => $plugin_settings_map[ $pay_method ],
							'ccexpiredate'  => $originalDate,
							'ccname'        => $first_name . " " . $last_name,
							'ccdefault'     => "T",
						];
					}
				} elseif ( !empty($ccData) ) {
					
					$originalDate = $ccData['date'];// mm/yy
					
					$card_type = $this->_cardType( $ccData['cardNumber'] );
					
					$pay_method = 'paymentmethod_' . $card_type;
					
					if ( $card_type !== 0 && ! empty( $plugin_settings_map[ $pay_method ] ) ) {
						//credit card array
						$this->credit_card_array = [
							'ccnumber'      => $ccData['cardNumber'],
							'paymentmethod' => $plugin_settings_map[ $pay_method ],
							'ccexpiredate'  => $originalDate,
							'ccname'        => $first_name . " " . $last_name,
							'ccdefault'     => "T",
						];
					}
				} elseif (
//Robert added to update default credit card.
					$order_pay_method == 'in8_netsuite_payments' ) {
					//credit card array
					$card_internal_id = get_post_meta( $order->get_id(), 'saved_card_id', TRUE );
					if( !empty($card_internal_id) ) {
					$this->credit_card_array = [
						'internalid' => $card_internal_id,
						'ccname'     => $first_name . " " . $last_name,
						'ccdefault'  => "T",
					];
					}
				}
			}
			if ( isset( $card_type ) ) {
				update_post_meta( $order->get_id(), '_card_type', $card_type );
			}
			
			if ( $order_pay_method == 'in8_netsuite_payments' && ( ! empty( $this->credit_card_array ) || ! empty( $_POST['select_credit_card'] ) ) ) {
//                $this->process_payment_on_ns = TRUE;
			}
			
			//get defined mapping fields
			$subsidiary = NULL;
			if ( empty( $this->customer_internal_id ) || ( ! empty( $this->customer_internal_id ) && ! isset( $plugin_settings_update_map['subsidiary'] ) ) ) {
				if ( ! empty( $plugin_settings_map['subsidiary'] ) ) {
					$subsidiary = $plugin_settings_map['subsidiary'];
				}
			}
			
			$taxable = NULL;
			if ( empty( $this->customer_internal_id ) || ( ! empty( $this->customer_internal_id ) && ! isset( $plugin_settings_update_map['taxable'] ) ) ) {
				if ( ! empty( $plugin_settings_map['taxable'] ) ) {
					$taxable = $plugin_settings_map['taxable'];
				}
			}
			
			$category = NULL;
			if ( empty( $this->customer_internal_id ) || ( ! empty( $this->customer_internal_id ) && ! isset( $plugin_settings_update_map['category'] ) ) ) {
				if ( ! empty( $plugin_settings_map['category'] ) ) {
					$category = $plugin_settings_map['category'];
				}
			}
			
			//define ship address
			$ship_fname   = $order->get_shipping_first_name();
			$ship_lname   = $order->get_shipping_last_name();
			$ship_company = $order->get_shipping_company();
			$shipaddr1    = $order->get_shipping_address_1();
			$shipaddr2    = $order->get_shipping_address_2();
			$shipcity     = $order->get_shipping_city();
			$shipstate    = $order->get_shipping_state();
			$shipcountry  = $order->get_shipping_country();
			$shipzip      = $order->get_shipping_postcode();
			
			//customer created date
			$datecreated = date( "Y-m-d H:i:s" );
			
			// If the shipping address is blank, copy from billing
			if ( empty( $ship_fname ) && empty( $ship_lname ) && empty( $shipaddr1 ) &&
			     empty( $shipaddr2 ) && empty( $shipcity ) && empty( $shipstate ) && empty( $shipcountry ) && empty( $shipzip )
			) {
				$ship_fname  = $order->get_billing_first_name();
				$ship_lname  = $order->get_billing_last_name();
				$ship_company = $order->get_billing_company();
				$shipaddr1   = $order->get_billing_address_1();
				$shipaddr2   = $order->get_billing_address_2();
				$shipcity    = $order->get_billing_city();
				$shipstate   = $order->get_billing_state();
				$shipcountry = $order->get_billing_country();
				$shipzip     = $order->get_billing_postcode();
			}
			
			//added by Robert- changed mapping below as well
			if ( ! empty( $bill_company ) ) {
				$billattention = $first_name . ' ' . $last_name;
			}
			if ( empty( $bill_company ) ) {
				$bill_company  = $first_name . ' ' . $last_name;
				$billattention = '';
			}
			if ( ! empty( $ship_company ) ) {
				$ship_attention = $ship_fname . ' ' . $ship_lname;
			}
			if ( empty( $ship_company ) ) {
				$ship_company   = $ship_fname . ' ' . $ship_lname;
				$ship_attention = '';
			}
//end add by Robert
			
			if ( isset( $customer_plugin_conf_options['new_address'] ) && $customer_plugin_conf_options['new_address'] == 'on' ) {
				$newaddress = "T";
			} else {
				$newaddress = "F";
			}
			
			
			// CyberSource Tokenization
			if ( $this->isCyberSourceTokenizationOn( $order ) ) {
				
				$card_types_alias = [
					'mc' => 'MasterCard',
					'amex' => 'American Express',
					'disc'  => 'Discover',
				];
				$token          = apply_filters('in8_so_token', get_post_meta( $order->get_id(), 'cybersource_token', TRUE ), $order);
				$card_type_name = apply_filters('in8_so_token_card_type', get_post_meta( $order->get_id(), 'cybersource_token_card_type', TRUE ), $order);
				$card_type_name = isset($card_types_alias[$card_type_name]) ? $card_types_alias[$card_type_name] : $card_type_name;
				$card_type_name = str_replace( ' ', '', $card_type_name );
				$pay_method     = strtolower( 'paymentmethod_' . $card_type_name );
				$expDate        = apply_filters('in8_so_token_expire_date', get_post_meta( $order->get_id(), 'cybersource_token_expire_date', TRUE ), $order);
				if( $order->get_payment_method() == 'cybersource_credit_card' ){
					$expDate = explode('-', $expDate);
					$expDate = $expDate[1] .'/20'.$expDate[0];
				}
				//credit card array
				if ( ! empty( $token ) && ! empty( $plugin_settings_map[ $pay_method ] ) ) {
					$this->credit_card_array = [
						'ccnumber'      => $token,
						'paymentmethod' => $plugin_settings_map[ $pay_method ],
						'ccexpiredate'  => $expDate,
						'ccmemo'        => 'Token',
//						'ccname'        => $first_name . " " . $last_name,
						'ccdefault'     => "T",
					];
				}
			}
			
			
			$customer = [
				'recordType'                    => "customer",
				'internalId'                    => $this->customer_internal_id,
				'custentity_in8_wc_customer_id' => $this->user_id,
				'isperson'                      => $is_person,
				'creditcards'                   => [ $this->credit_card_array ],
				'companyname'                   => $company,
				'category'                      => $category,
				'subsidiary'                    => $subsidiary,
				'taxable'                       => $taxable,
				'currency'                      => [
					$currency,
				], //required
				'customform'                    => $customform, //required
				'datecreated'                   => $datecreated, //required
				'email'                         => $customer_email,
				'firstname'                     => $first_name,
				'lastname'                      => $last_name,
				'phone'                         => $phone,
				//'addrphone'                     => $phone,
				'referrer'                      => "https://" . $_SERVER['HTTP_HOST'],
				'billlabel'                     => 'Billing',
				'billaddr1'                     => $address_1,
				'billaddr2'                     => $address_2,
				'billcity'                      => $city,
				'billstate'                     => $state,
				'billcountry'                   => $country,
				'billzip'                       => $postcode,
				'billphone'                     => $phone,
				'billattention'                 => $billattention, //changed by Robert
				'billaddressee'                 => $bill_company, //changed by Robert
				'shipattention'                 => $ship_attention,   //changed by Robert
				'shipaddressee'                 => $ship_company,  //changed by Robert
				'shiplabel'                     => 'Shipping',
				'shipaddr1'                     => $shipaddr1,
				'shipaddr2'                     => $shipaddr2,
				'shipcity'                      => $shipcity,
				'shipstate'                     => $shipstate,
				'shipcountry'                   => $shipcountry,
				'shipzip'                       => $shipzip,
				//'shipphone'                     => $phone,
				'newaddress'                    => $newaddress,
				'source'                        => "https://" . $_SERVER['HTTP_HOST'],
			];
			
			if ( empty( $this->credit_card_array ) ) {
				unset( $customer['creditcards'] );
			}
			
			if ( is_null( $currency ) ) {
				unset( $customer['currency'] );
			}
			
			$customer = array_filter( $customer );
			
			//conditional mapping for customers
			$this->conditionalMapping( $customer, $order, $this->customer_internal_id, 1 );
			
			//filter to allow editing of customer data before sync request
			$customer = apply_filters( 'in8_so_customer_after_prepared_data', $customer, $order );
			
			return $customer;
		}
		
		//Send order to Netsuite
		public function sendOrderToNetsuite( $order_id, $order_internal_id = 0, $is_sale = FALSE, $reset = FALSE ) {
			$this->shouldReset( $reset );
			
			
			if ( $this->doNotSync( $order_id, 'order' ) ) {
				$doNotSyncErrorMessage = apply_filters('in8_do_not_sync_error', false, $order_id, 'customer');
				if( !empty($doNotSyncErrorMessage) ){
					$netsuite_connection = new NetsuiteConnection( $order_id, 'salesorder-customer' );
					$netsuite_connection->processError( $doNotSyncErrorMessage );
				}
				return;
			}
			
			/**
			 * avoid extra calls on checkout if trying to send order once
			 */
			$is_checkout = is_page( wc_get_page_id( 'checkout' ) );
			
			$remove_hooks = apply_filters( 'in8_so_sync_remove_hooks', $is_checkout, $order_id );
			
			if ( $remove_hooks ) {
				$this->_removeOrderHooks();
			}
			
			if( apply_filters('in8_force_customer_sync_on_so', false, $order_id) ){
				$this->newOrderCreated( $order_id );
			}
			
			//this will update customer data on update request
			if ( $order_internal_id != 0 ) {
				$this->newOrderCreated( $order_id );
			} else {
				$order_internal_id = get_post_meta( $order_id, "internalid", TRUE );
				if ( $is_checkout && ! empty( $order_internal_id ) ) {
					
					$NS_auth_successful = get_post_meta( $order_id, "NS_auth_successful", TRUE );
					if ( $NS_auth_successful == 'no' ) {
					
					} else {
						return FALSE;
					}
				}
			}
			//return if a pruchase order
			if ( ! $is_sale ) {
				if ( $this->is_purchase_order || $this->isPurchaseOrder( $order_id ) ) {
					return FALSE;
				}
			}
			
			$this->order_id = $order_id;
			
			//get user associated with order
			$order_customer_id = get_post_meta( $order_id, '_customer_user', TRUE );
			
			if ( empty( $this->customer_internal_id ) ) {
				
				if ( $order_customer_id != 0 ) { //if a order have user assigned check for 'customerinternalid' in user meta
					$this->customer_internal_id = get_user_meta( $order_customer_id, 'customerinternalid', TRUE );
				}
				
				if ( empty( $this->customer_internal_id ) ) { //if 'customerinternalid' is empty, check in order meta for it
					$this->customer_internal_id = get_post_meta( $order_id, 'customerinternalid', TRUE );
				}
				
				if ( empty( $order_internal_id ) && empty( $this->customer_internal_id ) ) { //if 'customerinternalid' is empty, check in order meta for it
					$this->newOrderCreated( $order_id );
				}
			}
			
			
			//set user id to set zero in case admin is creating the order
			$this->user_id = $order_customer_id;
			
			if ( ! class_exists( 'NetsuiteConnection' ) ) { //in case of ajax
				//netsuite api connection file
				require_once( WP_PLUGIN_DIR . '/in8sync/code/netsuiteconnection.php' );
			}
			//netsuite connection object
			$netsuite_connection = new NetsuiteConnection( $this->order_id, 'salesorder' );
			
			//customer internal id is not set, then request returns from this point
			if ( empty( $this->customer_internal_id ) ) {
				$error = "Sales Order Error :: Customer Internal Id can't be found with customer or order meta fields.So NetSuite sales order can't be created.";
				$netsuite_connection->processError( $error );
				
				return FALSE; //return if customer internal ID is not defined
			}
			
			$order = new WC_Order( $this->order_id );
			
			//plugin options
			$plugin_options = In8_SalesOrder::getOptions();
			
			$order_data = $this->_getOrderData( $order, $plugin_options, $order_internal_id );
			
			if ( $order_data ) {
				
				if ( ! empty( $order_internal_id ) ) {
					$order_data['internalId'] = $order_internal_id;
				}
				
				if ( ! apply_filters( 'in8sync_so_order_should_sync', TRUE, $order_data, $order_id ) ) {
					return FALSE;
				}
				
				//create new sales order
				$order_response = $netsuite_connection->doRequest( $order_data );
				//make sure to remove all order hooks once order request is sent
				$this->_removeOrderHooks();
				
				if ( $order_response && isset( $order_response['data']['id'] ) ) {
					do_action( 'in8sync_so_order_after_sync', $this->order_id );
					
					update_post_meta( $this->order_id, "internalid", $order_response['data']['id'], TRUE );
					update_post_meta( $this->order_id, "salesordernumber", $order_response['data']['tranid'], TRUE );
					//update_post_meta($this->order_id, 'repeat_payment_test', '2');
					
					//means payment has been successfully processed on NS side
					if ( $this->process_payment_on_ns ) {
						if ( ! empty( $order_response['data']['paymentevent'] ) ) {
							$fix    = array_slice( $order_response['data']['paymentevent'], - 1 );
							$lastEl = array_pop( $fix );
							
							if ( $lastEl['result'] == 'Accept' ) {
								update_post_meta( $this->order_id, "netsuite_pnrefnum", $order_response['data']['pnrefnum'], TRUE );
								update_post_meta( $this->order_id, "NS_auth_successful", 'yes', TRUE );
								update_post_meta( $this->order_id, "_transaction_id", $order_response['data']['pnrefnum'], TRUE );
							} elseif ( $lastEl['result'] == 'Payment Hold' ) {
								update_post_meta( $this->order_id, 'NS_auth_error_message', 'Credit Card Authorization Failed, Please Try a Different Credit Card' );
								update_post_meta( $this->order_id, "NS_auth_successful", 'no', TRUE );
							} else {
								update_post_meta( $this->order_id, 'NS_auth_error_message', 'Unknown Credit Card Error' );
								update_post_meta( $this->order_id, "NS_auth_successful", 'no', TRUE );
							}
						} else {
							update_post_meta( $this->order_id, 'NS_auth_error_message', 'Error Connecting to Payment Gateway' );
							update_post_meta( $this->order_id, "NS_auth_successful", 'no', TRUE );
						}
					}
					
					if ( In8_SalesOrder::testFeature( 'enable_subscription' ) ) {
						$this->createNewSubscription( $this->order_id );
					}
				}
				
			}
		}
		
		public function _getOrderData( $order, $plugin_options, $order_internal_id = 0 ) {
			$getRequestDataExampleReq = isset( $_REQUEST['in8_get_order_json_request'] );
			
			//combine mapping keys and values and then change keys to lowercaSE
			$plugin_settings_map        = array_change_key_case( array_combine( $plugin_options['mapping_key'], $plugin_options['mapping_value'] ) );
			$plugin_settings_update_map = [];
			
			$co_options = In8_SalesOrder::getOptions( 2 );
			
			if ( $order_internal_id != 0 && ! empty( $plugin_options['mapping_update'] ) ) { //update
				$plugin_settings_update_map = array_combine( array_intersect_key( $plugin_options['mapping_key'], $plugin_options['mapping_update'] ), array_intersect_key( $plugin_options['mapping_update'], $plugin_options['mapping_key'] ) );
				$plugin_settings_update_map = array_change_key_case( $plugin_settings_update_map );
			}
			//plugin conf options
			$plugin_conf_options = In8_SalesOrder::getOptions( 2 );
			
			//get defined mapping fields
			
			$subsidiary = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['subsidiary'] ) ) ) {
				if ( ! empty( $plugin_settings_map['subsidiary'] ) ) {
					$subsidiary = $plugin_settings_map['subsidiary'];
				}
			}
			
			$customform = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['customform'] ) ) ) {
				if ( ! empty( $plugin_settings_map['customform'] ) ) {
					$customform = $plugin_settings_map['customform'];
				}
			}
			
			$terms = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['terms'] ) ) ) {
				if ( ! empty( $plugin_settings_map['terms'] ) ) {
					$terms = $plugin_settings_map['terms'];
				}
			}
			
			$location = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['location'] ) ) ) {
				if ( ! empty( $plugin_settings_map['location'] ) ) {
					$location = $plugin_settings_map['location'];
				}
			}
			
			$department = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['department'] ) ) ) {
				if ( ! empty( $plugin_settings_map['department'] ) ) {
					$department = $plugin_settings_map['department'];
				}
			}
			
			$opportunity = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['opportunity'] ) ) ) {
				if ( ! empty( $plugin_settings_map['opportunity'] ) ) {
					$opportunity = $plugin_settings_map['opportunity'];
				}
			}
			
			$leadsource = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['leadsource'] ) ) ) {
				if ( ! empty( $plugin_settings_map['leadsource'] ) ) {
					$leadsource = $plugin_settings_map['leadsource'];
				}
			}
			
			$salesrep = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['salesrep'] ) ) ) {
				if ( ! empty( $plugin_settings_map['salesrep'] ) ) {
					$salesrep = $plugin_settings_map['salesrep'];
				}
			}
			
			$partner = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['partner'] ) ) ) {
				if ( ! empty( $plugin_settings_map['partner'] ) ) {
					$partner = $plugin_settings_map['partner'];
				}
			}
			
			$class = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['class'] ) ) ) {
				if ( ! empty( $plugin_settings_map['class'] ) ) {
					$class = $plugin_settings_map['class'];
				}
			}
			
			$orderstatus = NULL;
			
			$woo_status = 'wc-'.$order->get_status();
			
			$search_status_index = ! empty( $plugin_conf_options['status_mapping_key'] ) ? array_search( $woo_status, $plugin_conf_options['status_mapping_key'] ) : FALSE;
			
			if ( $search_status_index !== FALSE ) {
				$orderstatus = $plugin_conf_options['status_mapping_value'] [ $search_status_index ];
			} elseif ( ! empty( $plugin_settings_map['orderstatus'] ) ) {
				$orderstatus = $plugin_settings_map['orderstatus'];
			}
			
			$istaxable = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['istaxable'] ) ) ) {
				if ( ! empty( $plugin_settings_map['istaxable'] ) ) {
					$istaxable = $plugin_settings_map['istaxable'];
				}
			}
			
			
			$total_taxes = NULL;
			if ( ! is_null( $istaxable ) ) {
				$total_taxes = $order->get_total_tax();
			}
			
			//get order subtotal
			$subtotal = $order->get_subtotal();
			
			$shipcomplete = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['shipcomplete'] ) ) ) {
				if ( ! empty( $plugin_settings_map['shipcomplete'] ) ) {
					$shipcomplete = $plugin_settings_map ['shipcomplete'];
				}
			}
			
			
			$shippingtaxcode = NULL;
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map['shippingtaxcode'] ) ) ) {
				if ( ! empty( $plugin_settings_map['shippingtaxcode'] ) ) {
					$shippingtaxcode = $plugin_settings_map['shippingtaxcode'];
				}
			}
			
			
			$currency = NULL;
			
			$order_currency = strtolower( get_woocommerce_currency() );
			
			// Get the currency applied to the order
			$on_order_currency = get_post_meta( $order->get_id(), '_order_currency', TRUE );
			
			// If the currency applied to order is not empty, use it
			$order_currency = ! empty( $on_order_currency ) ? strtolower( $on_order_currency ) : $order_currency;
			
			if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map [ 'currency_' . $order_currency ] ) ) ) {
				if ( isset( $plugin_conf_options['currency'] ) && $plugin_conf_options['currency'] == 'on' ) {
					//get woocommerce currency
					//check if order currency was deifned under settings tab
					if ( ! empty( $plugin_settings_map[ 'currency_' . $order_currency ] ) ) {
						//order currency
						$currency = [ "currency" => $plugin_settings_map[ 'currency_' . $order_currency ] ];
					}
				}
			}
			
			//order payment method
			$paymentmethod    = NULL;
			$is_credit_card   = FALSE;
			$order_pay_method = $order->get_payment_method();
			
			if ( $this->process_payment_on_ns ||
			     ! empty( $_POST[ $order_pay_method . '-card-number' ] ) ||
			     ! empty( $_POST['cardno'] ) ||
			     ! empty( $_POST[ $order_pay_method . '_accountNumber' ] ) ) {
				
				$card_num = ! empty( $_POST[ $order_pay_method . '-card-number' ] ) ? $_POST[ $order_pay_method . '-card-number' ] : '';
				if ( empty( $card_num ) ) {
					$card_num = ! empty( $_POST[ $order_pay_method . '_accountNumber' ] ) ? $_POST[ $order_pay_method . '_accountNumber' ] : '';
					if ( empty( $card_num ) ) {
						$card_num = ! empty( $_POST['cardno'] ) ? $_POST['cardno'] : '';
					}
				}
				
				//other pay method
				if ( ! empty( $card_num ) ) {
					//determine credit card type(visa, mastercard etc)
					$card_type = $this->_cardType( $card_num );
					
					//order pay method_getCouponInternalId
					$pay_method = 'paymentmethod_' . $card_type;
					if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map[ $pay_method ] ) ) ) {
						$paymentmethod = $plugin_settings_map[ $pay_method ];
					}
					$is_credit_card = TRUE;
				}
			} elseif ( ! empty( $_POST['select_credit_card'] ) && $order_pay_method == 'in8_netsuite_payments' ) { // for saved credit cards
				$is_credit_card = TRUE;
			}
			
			//other pay method
			$card_type = get_post_meta( $order->get_id(), '_card_type', TRUE );
			if ( $this->isCyberSourceTokenizationOn( $order ) ) {// if Cybersource Tokenization is on we use the credit card from there
				$card_type = strtolower( apply_filters('in8_so_token_card_type', get_post_meta( $order->get_id(), 'cybersource_token_card_type', TRUE ), $order) );
				$card_type = str_replace( ' ', '', $card_type );
			}
			if ( ! empty( $card_type ) ) {
				if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map[ 'paymentmethod_' . $card_type ] ) ) ) {
					if ( is_null( $paymentmethod ) && ! empty( $plugin_settings_map[ 'paymentmethod_' . $card_type ] ) ) {
						$paymentmethod = $plugin_settings_map[ 'paymentmethod_' . $card_type ];
					}
				}
			}
			if ( empty( $paymentmethod ) ) {
				if ( ! empty( $plugin_settings_map[ 'paymentmethod_' . $order_pay_method ] ) ) {
					$paymentmethod = $plugin_settings_map[ 'paymentmethod_' . $order_pay_method ];
				}
			}
			
			//order shipping method
			$shippingmethod = NULL;
			
			/* @var WC_Order $order */
			$shipping_methods = $order->get_shipping_methods();
			if ( ! empty( $shipping_methods ) ) {
				$method = array_shift( $shipping_methods );// get first in the array
				// combine the shipping method name and instance id if it exists like this: flat_rate:1
				$instance_id           = $method->get_instance_id();
				$instance_sub_id       = $method->get_meta( 'instance_sub_id', TRUE );
				$method_name           = $method->get_method_id() . ( ! empty( $instance_id ) ? ':' . $instance_id : '' ) . ( ! empty( $instance_sub_id ) ? ':' . $instance_sub_id : '' );
				$ship_method_data      = explode( ':', $method_name );
				$parent_shiping_method = $ship_method_data[0];
				if ( isset( $ship_method_data[1] ) ) {
					$child_shiping_method_id = $ship_method_data[1];
				}
				$sub_child_shipping_method_id = '';
				if ( isset( $ship_method_data[2] ) ) {
					$sub_child_shipping_method_id = ':' . trim( $ship_method_data[2] );
				}
				
				if ( strpos( trim( $parent_shiping_method ), 'table_rate' ) !== FALSE ) {
					
					$child_shiping_method = strtolower( 'shipmethod_table_rate-' . trim( $child_shiping_method_id ) . $sub_child_shipping_method_id );
					if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map[ $child_shiping_method ] ) ) ) {
						if ( array_key_exists( $child_shiping_method, $plugin_settings_map ) ) {
							$shippingmethod = $plugin_settings_map [ $child_shiping_method ];
						}
					}
				} elseif ( trim( $parent_shiping_method ) == 'ups' || trim( $parent_shiping_method ) == 'fedex' || trim( $parent_shiping_method ) == 'fedex_wsdl' ) {
					$child_shiping_method = strtolower( 'shipmethod_' . trim( $parent_shiping_method ) . "-" . trim( $child_shiping_method_id ) . $sub_child_shipping_method_id );
					if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map[ $child_shiping_method ] ) ) ) {
						if ( array_key_exists( $child_shiping_method, $plugin_settings_map ) ) {
							$shippingmethod = $plugin_settings_map [ $child_shiping_method ];
						}
					}
				}
				
				if ( is_null( $shippingmethod ) || empty( $shippingmethod ) ) {
					$order_shipping_method_id = strtolower( $method_name );
					if ( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map[ 'shipmethod_' . $order_shipping_method_id ] ) ) ) {
						
						if ( array_key_exists( 'shipmethod_' . $order_shipping_method_id, $plugin_settings_map ) ) {
							$shippingmethod = $plugin_settings_map [ 'shipmethod_' . $order_shipping_method_id ];
						}
					}
				}
			}
			
			// Default shipping method
			if ( is_null( $shippingmethod ) || empty( $shippingmethod ) ) {
				if( $order_internal_id == 0 || ( $order_internal_id != 0 && ! isset( $plugin_settings_update_map[ 'shipmethod_default' ] ) ) ) {
					if ( array_key_exists( 'shipmethod_default', $plugin_settings_map ) ) {
						$shippingmethod = $plugin_settings_map['shipmethod_default'];
					}
				}
			}
			
			$orderdate         = $order->get_date_created();
			$email             = $order->get_billing_email();
			$totalcostestimate = $order->get_total();
			
			$message = $order->get_customer_note();
			
			$status = $order->get_status();
			
			$createddate = $order->get_date_created()->date( 'd/m/Y' );
			
			$shippingcost = $order->get_total_shipping();
			
			//order coupon code for sending coupon's internal ID with order data to apply discounts
			$coupon_code_internal_id = NULL;
			
			// Send the coupon name to NetSuite
			$order_coupon_name = NULL;
			
			// Use this value to store the gift certificate id
			$gift_code_internal_id = false;
			
			if ( isset( $plugin_conf_options['enable_coupon'] ) && $plugin_conf_options['enable_coupon'] == 'on' ) {
				$order_coupons = $order->get_items( 'coupon' );
				
				if ( ! empty( $order_coupons ) ) {
					$order_coupon_rate = 0;
					foreach ( $order_coupons as $couponid => $order_coupon ) {
						/* @var WC_Order_Item_Coupon $order_coupon */
						$order_coupon_name = $order_coupon->get_name();
						$order_coupon_rate += $order_coupon->get_discount();
						$coupon            = new WC_Coupon( $order_coupon_name );
						
						//Robert Change 12/19
						$isgc = wc_get_coupon_id_by_code($order_coupon_name);
						if( !empty($isgc) ) {
							$isGiftCertificate = get_post_meta( $isgc, 'isGiftCertificate', TRUE );
							if ( ! empty( $isGiftCertificate ) && (bool) $isGiftCertificate ) {
								continue;
							}
						}
						//Robert End Change 12/19
						
						if ( ! empty( $coupon ) ) {
							$couponId                  = $coupon->get_id();
							$coupon_code_internal_id   = get_post_meta( $couponId, 'internalId', TRUE );
							$coupon_code_discount_item = get_post_meta( $couponId, 'discountitem', TRUE );
						}
					}
				}
			}
			
			// Gift Certificate Or Voucher Mapping
			if (In8_SalesOrder::testFeature( 'use_gift_certificate' )) {
				if( empty($order_coupons) ){
					$order_coupons = $order->get_items('coupon');
				}
				if (!empty($order_coupons)) {
					foreach ($order_coupons as $order_coupon) {
						/* @var WC_Order_Item_Coupon  $order_coupon */
						$couponId = wc_get_coupon_id_by_code($order_coupon->get_name());
						if( !empty($couponId) ) {
							$isGiftCertificate = get_post_meta( $couponId, 'isGiftCertificate', TRUE );
							if ( ! empty( $isGiftCertificate ) && (bool) $isGiftCertificate ) {
								$gift_code_internal_id = get_post_meta( $couponId, 'internalId', TRUE );
							}
						}else{
							$voucher_post      = get_page_by_title( $order_coupon->get_code(), OBJECT, 'wc_voucher' );
							if( !empty($voucher_post) ){
								$voucher_id        = $voucher_post->ID;
								$isGiftCertificate = get_post_meta( $voucher_id, 'isGiftCertificate', TRUE );
								if ( ! empty( $isGiftCertificate ) && (bool) $isGiftCertificate ) {
									$gift_code_internal_id = get_post_meta( $voucher_id, 'internalId', TRUE );
								}
							}
						}
					}
				}
			}
			
			// Checks if the order should be sent as Estimate to NetSuite
			$isEstimate = NULL;
			
			if ( isset( $plugin_conf_options['sync_estimate'] ) && $plugin_conf_options['sync_estimate'] == 'on' ) {
				$isEstimate = 'T';
			}
			
			//order line items
			$order_items = $order->get_items();
			
			$items = [];
			$i     = 0;
			
			$per_item_pricing = 0;
			
			$missing_internal_id = [];
			foreach ( $order_items as $order_item_key => $order_item ) {
				/* @var WC_Order_Item_Product $order_item */
				if ( In8_SalesOrder::testFeature( 'enable_product_bundles' ) ) {
					// This script will make sure that when using the bundle item feature it will only sync the parent/kit item to NetSuite
					if ( isset( $order_item['bundled_items'] ) && isset( $order_item['per_product_pricing'] ) && $order_item['per_product_pricing'] == 'yes' ) {
						$per_item_pricing = 1;
					} elseif ( isset( $order_item['bundled_item_id'] ) ) {
						if ( $per_item_pricing == 0 ) {
							continue;
						}
					} else {
						$per_item_pricing = 0;
					}
				}
				
				if ( $order_item['variation_id'] != 0 ) {
					$internalid = get_post_meta( $order_item['variation_id'], 'internalid', TRUE );
					if ( empty( $internalid ) ) {
						$internalid = get_post_meta( $order_item['product_id'], 'internalid', TRUE );
					}
				} else {
					$internalid = get_post_meta( $order_item['product_id'], 'internalid', TRUE );
				}
				
				// If the internalId is missing
				// ~ Check if the genericItemId is set and if this item is connected to the GI
				// ~ And collect the internalId from the item meta
				$genericItemId = In8_Customers_Sales_Options::getGenericItem();
				if ( empty( $internalid ) && ! empty( $genericItemId ) && $order_item->get_product_id() == $genericItemId ) {
					$internalid = $order_item->get_meta( 'internalid', TRUE );
				}
				
				//get order line items netsuite internal id
				if ( ! empty( $internalid ) ) {
					
					// check if the item is storing a gravity form entry
					if( in8_is_plugin_active('woocommerce-gravityforms-product-addons/gravityforms-product-addons.php') ) {
						$tempData = $order_item->get_meta( '_gravity_forms_history', TRUE );
						if ( ! empty( $tempData ) ) {
							$this->entry_id = isset( $tempData['_gravity_form_linked_entry_id'] ) ? $tempData['_gravity_form_linked_entry_id'] : NULL;
						}
					}
					
					$items[ $i ] = [ 'item' => $internalid, 'quantity' => (float) $order_item['qty'] ];
					
					if ( ! is_null( $coupon_code_internal_id ) || ! is_null( $order_coupon_name ) ) {
						$items[ $i ]['amount'] = (float) $order_item['line_subtotal'];
						$items[ $i ]['rate']   = $order->get_item_subtotal( $order_item, FALSE, FALSE );
					} else {
						$items[ $i ]['amount'] = (float) $order_item['line_total'];
						$items[ $i ]['rate']   = $order->get_item_total( $order_item, FALSE, FALSE );
					}
					
					// Default price level to -1 but have the possibility to map it from the SO options
					$price_level = -1;
					$mapped_price_level = In8_SalesOrder::getOptionValue( 'price_level', $order_internal_id );
					if( !empty($mapped_price_level) ){
						$price_level = $mapped_price_level;
					}
					$items[ $i ]['price'] = $price_level;
					
					// Gift Certificate OR Voucher Mapping
					if( In8_SalesOrder::testFeature( 'use_gift_certificate' ) ) {
						if ( $order_item['product_id'] ) {
							$sku = get_post_meta( $order_item['product_id'], '_sku', TRUE );
							if ( in_array($sku, apply_filters('in8_gift_skus', ['giftcertificate', 'giftvoucher'])) ) {
								$wccoupons = get_post_meta( $order->get_id(), 'sc_coupon_receiver_details', TRUE );
								if ( ! empty( $wccoupons ) ) {
									foreach ( $wccoupons as $wccoupon ) {
										$giftcertfrom           = $wccoupon['code'];
										$giftcertrecipientemail = 'giftcertificate@in8sync.com';
									}
									$wccodeid              = get_page_by_title( $giftcertfrom, OBJECT, 'shop_coupon' );
									$giftcertrecipientname = $wccodeid->ID;
								} else {
									$voucher_id = $order_item->get_meta( '_voucher_id', TRUE );
									if ( ! empty( $voucher_id ) ) {
										$voucher_post           = get_post( $voucher_id );
										$giftcertrecipientname  = $voucher_id;
										$giftcertfrom           = $voucher_post->post_title;
										$giftcertrecipientemail = 'giftcertificate@in8sync.com';
									}
								}
								
								if ( !empty($giftcertrecipientname) ) {
									$items[ $i ]['giftcertrecipientname']  = $giftcertrecipientname;
									$items[ $i ]['giftcertrecipientemail'] = $giftcertrecipientemail;
									$items[ $i ]['giftcertfrom']           = $giftcertfrom;
								}
							}
						}
					}
					
					// allows editing of line items
					$items[ $i ] = apply_filters( 'in8_so_sync_line_item', $items[ $i ], $order->get_id(), $order_item_key, $order_item, $coupon_code_internal_id );
				} else {
					$missing_internal_id[] = $order_item['product_id'];
				}
				$i ++;
			}
			
			if ( ! $getRequestDataExampleReq ) {
				if ( ! empty( apply_filters( 'in8sync_so_missing_internal_ids', $missing_internal_id, $order ) ) ) {
					$netsuite_connection = new NetsuiteConnection( $this->order_id, 'salesorder' );
					$netsuite_connection->processError( "This item" . ( count( $missing_internal_id ) > 1 ? 's are' : ' is' ) . " missing the internal id: " . implode( $missing_internal_id, ',' ) );
					
					return;
				}
			}
			
			// Allow overriding the line items
			$items = apply_filters( 'in8_so_sync_line_items', $items, $order->get_id(), $order_items );
			
			// Add Fee Items
			$order_items = $order->get_items( 'fee' );
			
			foreach ( $order_items as $order_item ) {
				if ( isset( $plugin_settings_map[ 'fee:' . strtolower( $order_item->get_name() ) ] ) ) {
					$fee_item = $plugin_settings_map[ 'fee:' . strtolower( $order_item->get_name() ) ];
					
					$items[ $i ]         = [ 'item' => $fee_item, 'quantity' => 1 ];
					$items[ $i ]['rate'] = $order->get_item_total( $order_item, FALSE, FALSE );
					$i ++;
					continue;
				}
				
				// Get the Fee Internal Id from the Sales Orders mappings page (key "fee")
				if ( array_key_exists( 'fee', $plugin_settings_map ) ) {
					$fee_item = $plugin_settings_map['fee'];
					
					$items[ $i ]         = [ 'item' => $fee_item, 'quantity' => 1 ];
					$items[ $i ]['rate'] = $order->get_item_total( $order_item, FALSE, FALSE );
					$i ++;
				}
			}
			
			//For sending discount item
			$discountitem = NULL;
			$discountrate = NULL;
			if ( isset( $plugin_conf_options['enable_discountitem'] ) && $plugin_conf_options['enable_discountitem'] == 'on' ) {
				if ( isset( $plugin_conf_options['enable_sendline'] ) && $plugin_conf_options['enable_sendline'] == 'on' ) {
					if( !empty($coupon_code_discount_item) ) {
						$items[]['internalid'] = $coupon_code_discount_item;
					}
				} else {
					$discountrate = -$order_coupon_rate;
					$discountitem = $coupon_code_discount_item;
				}
			}
			
			$record_type = "salesorder";
			
			
			$bill_fname  = $order->get_billing_first_name();
			$bill_lname  = $order->get_billing_last_name();
			$billcountry = $order->get_billing_country();
			$billphone   = $order->get_billing_phone();
			$billaddr1   = $order->get_billing_address_1();
			$billaddr2   = $order->get_billing_address_2();
			$billcity    = $order->get_billing_city();
			$billstate   = $order->get_billing_state();
			$billzip     = $order->get_billing_postcode();
			$bill_company = $order->get_billing_company();
			
			//define ship address
			$ship_fname   = $order->get_shipping_first_name();
			$ship_lname   = $order->get_shipping_last_name();
			$ship_company = $order->get_shipping_company();
			$shipaddr1    = $order->get_shipping_address_1();
			$shipaddr2    = $order->get_shipping_address_2();
			$shipcity     = $order->get_shipping_city();
			$shipstate    = $order->get_shipping_state();
			$shipcountry  = $order->get_shipping_country();
			$shipzip      = $order->get_shipping_postcode();
			
			// If the shipping address is blank, copy from billing
			if ( empty( $ship_fname ) && empty( $ship_lname ) && empty( $shipaddr1 ) &&
			     empty( $shipaddr2 ) && empty( $shipcity ) && empty( $shipstate ) && empty( $shipcountry ) && empty( $shipzip )
			) {
				$ship_fname  = $order->get_billing_first_name();
				$ship_lname  = $order->get_billing_last_name();
				$ship_company  = $order->get_billing_company();
				$shipaddr1   = $order->get_billing_address_1();
				$shipaddr2   = $order->get_billing_address_2();
				$shipcity    = $order->get_billing_city();
				$shipstate   = $order->get_billing_state();
				$shipcountry = $order->get_billing_country();
				$shipzip     = $order->get_billing_postcode();
			}
			
			//added by Robert- changed mapping below as well
			if ( ! empty( $bill_company ) ) {
				$billattention = $bill_fname . ' ' . $bill_lname;
			}
			if ( empty( $bill_company ) ) {
				$bill_company  = $bill_fname . ' ' . $bill_lname;
				$billattention = '';
			}
			if ( ! empty( $ship_company ) ) {
				$ship_attention = $ship_fname . ' ' . $ship_lname;
			}
			if ( empty( $ship_company ) ) {
				$ship_company   = $ship_fname . ' ' . $ship_lname;
				$ship_attention = '';
			}
//end add by Robert
			//get user associated with order
			$customerNsId = $this->customer_internal_id;
			if ( ! $getRequestDataExampleReq ) { // stop it from returning the wrong customer internal id when retrieving the order json request
				$checkForWrongCustomerId = $this->check_for_wrong_internal_id( $order );
				if ( ! empty( $checkForWrongCustomerId ) ) {
					$customerNsId = $checkForWrongCustomerId;
				}
			}
			
			$order_data = [
				'recordType'               => $record_type,
				'custbody_in8_wc_order_id' => $this->order_id,
				'currency'                 => $currency['currency'],
				'entity'                   => $customerNsId,
				'status'                   => $status,
				'shippingcost'             => $shippingcost,
				'overrideshippingcost'     => $shippingcost,
				'email'                    => $email,
				'createddate'              => $createddate,
				'subsidiary'               => $subsidiary,
				'customform'               => $customform,
				'terms'                    => $terms,
				'location'                 => $location,
				'department'               => $department,
				'opportunity'              => $opportunity,
				'leadsource'               => $leadsource,
				'salesrep'                 => $salesrep,
				'partner'                  => $partner,
				'class'                    => $class,
				'orderstatus'              => $orderstatus,
				'istaxable'                => $istaxable,
				'taxtotal'                 => $total_taxes,
				'taxamountoverride'        => $total_taxes,
				'subtotal'                 => $subtotal,
				'shipcomplete'             => $shipcomplete,
				'message'                  => $message,
				'source'                   => "https://" . $_SERVER['HTTP_HOST'],
				'totalcostestimate'        => $totalcostestimate,
				'shipmethod'               => $shippingmethod,
				'shippingtaxcode'          => $shippingtaxcode,
				'item'                     => $items,
				'discountitem'             => $discountitem,
				'discountrate'             => $discountrate,
				'billattention'            => $billattention,  //Changed by Robert
				'shipattention'            => $ship_attention, //Changed by Robert
				'billaddressee'            => $bill_company, //Changed by Robert
				'shipaddressee'            => $ship_company, //Changed by Robert
				'firstname'                => $order->get_billing_first_name(),
				'lastname'                 => $order->get_billing_last_name(),
				'paymentmethod'            => $paymentmethod,
				'couponcode'               => $coupon_code_internal_id,
				'isestimate'               => $isEstimate,
				'billcountry'              => $billcountry,
				'billphone'                => $billphone,
				'billaddr1'                => $billaddr1,
				'billaddr2'                => $billaddr2,
				'billcity'                 => $billcity,
				'billstate'                => $billstate,
				'billzip'                  => $billzip,
				'shipcountry'              => $shipcountry,
				'shipaddr1'                => $shipaddr1,
				'shipaddr2'                => $shipaddr2,
				'shipcity'                 => $shipcity,
				'shipstate'                => $shipstate,
				'shipzip'                  => $shipzip,
			];
			
			if ( isset( $plugin_conf_options['enable_coupon'] ) && $plugin_conf_options['enable_coupon'] == 'on' && $plugin_conf_options['coupon_text'] == 'on' ) {
				unset( $order_data['couponcode'] );
				$order_data['couponcode.text'] = $order_coupon_name;
			}
			
			// Map the gift certificate id if the order has one
			if( In8_SalesOrder::testFeature( 'use_gift_certificate' ) ) {
				if ( ! empty( $gift_code_internal_id ) ) {
					$order_data['giftcertificate'] = $gift_code_internal_id;
					
					//Robert Change 12/19
					//$order_data['couponcode']      = '';
					//$order_data['couponcode.text'] = '';
					//Robert End Change 12/19
				}
			}
			
			//only enable NS payment processing for credit cards
			if ( $is_credit_card ) {
				// if payment has to be processed on Netsuite side
				if ( $this->process_payment_on_ns ) {
					// If the option to get authorization in NetSuite is selected
					if ( $co_options['sync_get_auth'] == 'on' ) {
						$order_data['getauth'] = "T";
						if ( ! empty( $_POST['select_credit_card'] ) && $_POST['select_credit_card'] != 'new' ) {
							$order_data['creditcard'] = $_POST['select_credit_card'];
						}
						
						$auth_successful = get_post_meta( $order->get_id(), "NS_auth_successful", TRUE );
						if ( $auth_successful = 'no' ) {
							$ns_ccinternalid = get_post_meta( $order->get_id(), 'ns_ccinternalid', TRUE );
							if ( ! empty( $ns_ccinternalid ) ) {
								$order_data['creditcard'] = $ns_ccinternalid;
							}
						}
					}
				} else {
					$order_data['ccapproved'] = "T";
					$pay_tran_id              = get_post_meta( $this->order_id, "_transaction_id", TRUE );
					if ( empty( $pay_tran_id ) ) {
						$pay_tran_id = get_post_meta( $this->order_id, $this->transactionIds[ $order_pay_method ], TRUE );
						if ( ! empty( $pay_tran_id ) ) {
							$order_data['pnrefnum'] = $pay_tran_id;
						}
					} else {
						$order_data['pnrefnum'] = $pay_tran_id;
					}
				}
			}
			
			// map deposit id to SO
			if( !empty($order->get_meta('deposit_id', true)) ){
				$order_data['custbody_in8_deposit'] = $order->get_meta('deposit_id', true);
			}
			
			//array filter is used to remove empty values
			$order_data = array_filter( $order_data );
			
			//apply conditional mapping
			$this->conditionalMapping( $order_data, $order, $order_internal_id );
			
			if ( $shippingcost == 0 ) {
				$order_data['overrideshippingcost'] = (float) $shippingcost;
				$order_data['shippingcost']         = (float) $shippingcost;
			}
			
			// Contact Sync - Runs after conditional mapping
			if ( $this->contact_id && $this->customer_internal_id ) {
				$order_data['entity'] = $this->customer_internal_id;
			}
			
			$order_customerinternalid = get_post_meta( $order->get_id(), 'customerinternalid', TRUE );
			$customer_id              = $order->get_customer_id();
			if ( empty( $customer_id ) && ! empty( $order_customerinternalid ) ) {
				$this->customer_internal_id = $order_customerinternalid;
				$order_data['entity']       = $this->customer_internal_id;
			}
			
			// filter to allow data to be altered before it is send to NS
			$order_data = apply_filters( 'in8sync_so_order_after_data_prepared', $order_data, $order->get_id() );
			
			// Move the discount rate to always be at the end
			if( !empty($order_data['discountrate']) ){
				$temp_discountrate = $order_data['discountrate'];
				unset($order_data['discountrate']);
				$order_data['discountrate'] = $temp_discountrate;
			}
			
			
			return $order_data;
		}
		
		public function shouldReset( $reset ) {
			if ( $reset ) {
				$this->user_id              = 0;
				$this->order_id             = 0;
				$this->customer_internal_id = 0;
				$this->contact_id           = 0;
				$this->credit_card_array    = [];
			}
		}
		
		/**
		 * This function will check if the customer internal id from the sync does not match the one in the user meta
		 * If it does not it will use the one from the user meta and send a notice
		 *
		 * @param WC_Order $order
		 *
		 * @return integer
		 */
		private function check_for_wrong_internal_id( $order ) {
			// TODO - create a setting, that is default to on
			if ( apply_filters( 'in8_disable_check_for_wrong_internal_id', FALSE, $order ) ) {
				return NULL;
			}
			$customerNsId  = $this->customer_internal_id;
			$customer_id   = get_post_meta( $order->get_id(), '_customer_user', TRUE );
			//check if user is guest
			if( empty($customer_id) ) return $customerNsId;
			
			$checkCustomer = get_user_meta( $customer_id, 'customerinternalid', TRUE );
			
			// if customerNsId is empty, customer was not yet synced
			if( empty($customerNsId) ){
				return $checkCustomer;
			}
			
			// if not empty but different fix them
			if ( $checkCustomer != $customerNsId ) {
					$customerNsId      = $checkCustomer;
					
					update_post_meta( $this->order_id, 'customerinternalid', $customerNsId );
					update_post_meta( $this->order_id, 'wrongcustomerinternalid', $this->customer_internal_id );
					// netsuite connection object
						$netsuite_connection = new NetsuiteConnection( $this->order_id, 'salesorder' );
						$error               = "DATA ERROR - Wrong Customer Internal ID";
						$netsuite_connection->processError( $error );
				
			}
			
			return $customerNsId;
		}
		
		
		private function conditionalMapping( &$data, $order = NULL, $internal_id = 0, $mapping_type = 0 ) {
			//conditional mapping plugin options
			if ( $mapping_type == 1 ) {
				$cm_options = In8_Customer::getOptions( 1 );
			} else {
				$cm_options = In8_SalesOrder::getOptions( 1 );
			}
			
			if ( $this->user_id == 0 ) {
				$this->user_id = $order->get_customer_id();
			}
			
			if ( ! empty( $cm_options ) ) {
				foreach ( $cm_options as $mapping ) {
					if ( is_null( $order ) && ( $mapping['type'] == 3 || $mapping['type'] == 4 ) ) {
						continue;
					}
					
					if ( $this->user_id == 0 && ( $mapping['type'] == 1 || $mapping['type'] == 2 ) ) {
						continue;
					}
					
					if ( $internal_id !== 0 && isset( $mapping['exlcude_in_update'] ) ) { //skip if admin wants to exlcude this update request
						continue;
					}
					if ( ! empty( $mapping['ns_field_key'] ) ) {
						$ns_field_key = trim( $mapping['ns_field_key'] );
					} else {
						continue;
					}
					
					
					// GRAB THE ENTRY ID FROM THE LINE ITEM
					$entry_id = apply_filters('in8_so_entry_id', $this->entry_id);
					
					$mapped_value = ConditionalMapping::ConditionalMappingOperators( $mapping, $order, $this->user_id,$entry_id  );
					if ( ! empty( $mapped_value ) ) {
						$data[ $ns_field_key ] = $mapped_value;
					}
				}
			}
			
			return;
		}

//determines the credit card type
		private function _cardType( $number ) {
			$number = preg_replace( '/[^\d]/', '', $number );
			if ( preg_match( '/^3[47][0-9]{13}$/', $number ) ) {
				return 'americanexpress';
			} elseif ( preg_match( '/^3(?:0[0-5]|[68][0-9])[0-9]{11}$/', $number ) ) {
				return 'dinersclub';
			} elseif ( preg_match( '/^6(?:011|5[0-9][0-9])[0-9]{12}$/', $number ) ) {
				return 'discover';
			} elseif ( preg_match( '/^(?:2131|1800|35\d{3})\d{11}$/', $number ) ) {
				return 'jcb';
			} elseif ( preg_match( '/^5[1-5][0-9]{14}$/', $number ) ) {
				return 'mastercard';
			} elseif ( preg_match( '/^4[0-9]{12}(?:[0-9]{3})?$/', $number ) ) {
				return 'visa';
			} else {
				return 0;
			}
		}
		
		public function createNewSubscription( $order_id ) {
			$order   = wc_get_order( $order_id );
			$user_id = $order->get_customer_id();
			
			$is_subscription = FALSE;
			$items           = $order->get_items();
			foreach ( $items as $item ) {
				// check if the order has subscription items
				/* @var WC_Order_Item_Product $item */
				$auto_delivery_options = get_post_meta( $item->get_product_id(), 'auto_delivery_options', TRUE );
				if ( ! empty( $auto_delivery_options ) ) {
					$is_subscription = TRUE;
				}
			}
			// order does not have subscription items
			if ( ! $is_subscription ) {
				return;
			}
			
			
			$paymentmethod = apply_filters('in8_so_token_card_type', get_post_meta( $order->get_id(), 'cybersource_token_card_type', TRUE ), $order);
			
			//START customization to check if payment method is American Express, then strip/take out the space. TODO let Daniel know about this issue so he can update base plugin
			if ( $paymentmethod == 'American Express' ) {
				$paymentmethod = 'AmericanExpress';
			}
			//END customization
			
			$billingaddress  = get_post_meta( $order_id, 'billing_address_id', TRUE );
			$shippingaddress = get_post_meta( $order_id, 'shipping_address_id', TRUE );
			// it's a guest order
			if ( empty( $user_id ) ) {
				$billingaddress  = empty( $billingaddress ) ? 'checkout_data' : $billingaddress;
				$shippingaddress = empty( $shippingaddress ) ? 'checkout_data' : $shippingaddress;
			} else {
				$billingaddress  = empty( $billingaddress ) ? 'last_used_billing' : $billingaddress;
				$shippingaddress = empty( $shippingaddress ) ? 'last_used_shipping' : $shippingaddress;
			}
			
			$subscriptions = [];
			// find the subscription items
			foreach ( $order->get_items() as $item ) {
				$recurrence = $item->get_meta( 'subscription_options' );
				if ( empty( $recurrence ) ) {
					continue;
				}
				
				$item_product_id                = $item->get_product_id();
				$subscriptions[ $recurrence ][] = [
					'product_id' => $item_product_id,
					'quantity'   => $item->get_quantity(),
				];
			}
			
			if ( ! empty( $subscriptions ) ) {
				foreach ( $subscriptions as $recurrence => $items ) {
					$subscription_id = get_post_meta( $order_id, 'subscription_id_' . $recurrence, TRUE );
					// only sync the subscription from the order once, otherwise we will lose the data that was modified on the portal
					if ( ! empty( $subscription_id ) ) {
						continue;
					}
					
					$this->syncSubscription( $order_id, $subscription_id, $user_id, $billingaddress, $shippingaddress, $paymentmethod, $recurrence, $items );
				}
			}
			
		}
		
		/**
		 * @param       $order_id
		 * @param       $id
		 * @param       $user_id
		 * @param       $billingaddress_id
		 * @param       $shippingaddress_id
		 * @param       $paymentmethod
		 * @param       $recurrence_val
		 * @param       $items
		 * @param array $opts
		 */
		public function syncSubscription( $order_id, $id, $user_id, $billingaddress_id, $shippingaddress_id, $paymentmethod, $recurrence_val, $items, $opts = [] ) {
			$netsuite_connection = new NetsuiteConnection( $user_id, 'customer' );
			
			$customer_data = [];
			
			$address_types     = [ 'shipping', 'billing' ];
			$wcma_customer     = new WCMCA_Customer();
			$bill_address_data = [];
			$ship_address_data = [];
			foreach ( $address_types as $type ) {
				$address_id = $type == 'shipping' ? $shippingaddress_id : $billingaddress_id;
				if ( empty( $address_id ) && $address_id != 0 ) {
					$address_id = $type == 'shipping' ? 'last_used_shipping' : 'last_used_billing';
				}
				$address_data = $this->get_address_by_id( $user_id, $address_id, $order_id );
				if ( $type == 'shipping' ) {
					$ship_address_data = $address_data;
				} else {
					$bill_address_data = $address_data;
				}
				
				if ( ! isset( $address_data['internalId'] ) ) {
					$customer_data = $this->get_customer_data_with_address( $address_id, $address_data, $customer_data );
				} else {
					if ( $type == 'billing' ) {
						$billingaddress = $address_data['internalId'];
					} else {
						$shippingaddress = $address_data['internalId'];
					}
					
				}
			}
			
			$customer_internal_id = $this->customer_internal_id;
			if ( empty( $customer_internal_id ) ) {
				$customer_internal_id = $this->get_customer_internal_id( $order_id, $user_id );
			}
			
			if ( ! empty( $customer_data ) && ! empty( $customer_internal_id ) ) {
				$customer_data     = array_merge( [
					'recordType' => 'customer',
					'internalId' => $customer_internal_id,
				], $customer_data );
				$customer_data     = array_filter( $customer_data );
				$customer_response = $netsuite_connection->doRequest( $customer_data );
				if ( $customer_response && isset( $customer_response['addresses'] ) ) {
					// save address id
					if ( empty( $customer_response['addresses'] ) ) {
						$netsuite_connection->processError( "Addresses were not returned from ns" );
						
						return;
					}
					foreach ( $customer_response['addresses'] as $adr ) {
						if ( $adr['bill_multi_address_id'] == $billingaddress_id ) {
							$bill_address_data['internalId'] = $adr['internalid'];
							$billingaddress                  = $adr['internalid'];
							if ( ! empty( $user_id ) ) {
								$wcma_customer->update_addresses( $user_id, $billingaddress_id, $bill_address_data );
							}
						}
						if ( $adr['ship_multi_address_id'] == $shippingaddress_id ) {
							$ship_address_data['internalId'] = $adr['internalid'];
							$shippingaddress                 = $adr['internalid'];
							if ( ! empty( $user_id ) ) {
								$wcma_customer->update_addresses( $user_id, $shippingaddress_id, $ship_address_data );
							}
						}
					}
				} else {
					$netsuite_connection->processError( "Failed saving address information to ns" );
					
					return;
				}
			}
			
			
			// sync the subscription
			$netsuite_connection = new NetsuiteConnection( $user_id, 'subscription' );
			
			$paymentmethod_id = 0;
			if ( is_numeric( $paymentmethod ) ) {
				$paymentmethod_id = $paymentmethod;
			} else if ( $this->get_payment_method_internal_id( $paymentmethod ) ) {
				$paymentmethod_id = $this->get_payment_method_internal_id( $paymentmethod );

			}
			
			// get the credit card token
			$credit_card_token = '';
			if( isset($opts['token_id']) ) {
				$token_id = $opts['token_id'];
				if( !empty($token_id) ) {
					$token             = new WC_Payment_Token_CC( $token_id );
					$credit_card_token = $token->get_token();
				}
			}else{
				$credit_card_token = get_post_meta($order_id, 'cybersource_token', true);
			}
			
			$start       = date( 'm/d/Y' );
			$end         = date( 'm/d/Y', strtotime( '+1 year', time() ) );
			$nextPayment = 0;// TODO - how is this calculated
			$recurrence  = strtolower( $recurrence_val );
			switch ( $recurrence ) {
				case 'single event':
					$nextPayment = NULL;
					break;
				case 'monthly':
					$nextPayment = date( 'm/d/y', strtotime( "+1 month", time() ) );
					break;
				case 'annual':
					$nextPayment = date( 'm/d/y', strtotime( "+1 year", time() ) );
					break;
				case '3 months':
					$nextPayment = date( 'm/d/y', strtotime( "+3 months", time() ) );
					break;
				case '6 months':
					$nextPayment = date( 'm/d/y', strtotime( "+6 months", time() ) );
					break;
			}
			// if we're sending the date on the opts array replace it
			if ( isset( $opts['nextPayment'] ) ) {
				$nextPayment = $opts['nextPayment'];
			}
			
			$recurrence_options = $this->getSubscriptionRecurrenceOptions();
			$recurrence_id      = $recurrence_options[ $recurrence ];
			if ( empty( $recurrence_id ) ) {
				$netsuite_connection->processError( "ERROR :: Reccurrence does not have a accepted value, check the mapping or the value that is being sent." );
				
				return;
			}
			
			$custrecord_in8subs_active        = 'T';
			$custrecord_in8subs_bill_order    = 'F';
			$custrecord_in8subs_fulfill_order = 'F';
			$custrecord_in8subs_get_auth      = 'F';
			
			$sub_parent = [
				'custrecord_in8subs_active'           => $custrecord_in8subs_active,
				'custrecord_in8subs_bill_order'       => $custrecord_in8subs_bill_order,
				'custrecord_in8subs_fulfill_order'    => $custrecord_in8subs_fulfill_order,
				'custrecord_in8subs_get_auth'         => $custrecord_in8subs_get_auth,
				'custrecord_in8subs_billing_address'  => $billingaddress,
				'custrecord_in8subs_shipping_address' => $shippingaddress,
				'custrecord_in8subs_nextpayment'      => $nextPayment,
				'custrecord_in8subs_payment_method'   => $paymentmethod_id,
				'custrecord_in8subs_recurrence'       => $recurrence_id,
			];
			if( !empty($credit_card_token) ){
				$sub_parent['custrecord_in8_sub_creditcard_token'] = $credit_card_token;
			}
			
			if ( ! empty( $id ) ) {
				$sub_parent['id'] = $id;
			}
			$subscription_request = [
				'recordType'                       => 'customer',
				'internalId'                       => $customer_internal_id,
				'recmachcustrecord_in8subs_parent' => [
					$sub_parent,
				],
			];
			$subscription_request = apply_filters( 'in8_so_sync_subscription_request_data', $subscription_request );
			
			$subscription_response = $netsuite_connection->doRequest( $subscription_request );
			if ( empty( $subscription_response['data'] ) || $subscription_response['success'] != 1 ) {
				$netsuite_connection->processError( "Failed saving the subscription" );
				
				return;
			} else {
				$data     = $subscription_response['data'];
				$subs     = $data['recmachcustrecord_in8subs_parent'];
				$this_sub = end( $subs );
				
				// save line items request
				if ( empty( $id ) ) {
					$id = $this_sub['id'];
					update_post_meta( $order_id, 'subscription_id_' . $recurrence_val, $id );
				}
				
				$_items = [];
				foreach ( $items as $item ) {
					/* @var WC_Product $product */
					$product     = wc_get_product( $item['product_id'] );
					if ( is_wp_error( $product ) || empty( $product ) ) {
						continue;
					}
					
					$internal_id = get_post_meta( $product->get_id(), 'internalid', TRUE );
					$qty         = $item['quantity'];
					if ( empty( $internal_id ) ) {
						$netsuite_connection->processError( "Product {$product->get_id()} doesn't have an internalId, it can't be sent with the subscription." );
						continue;
					}
					
					$data       = [
						"custrecord_in8subs_item_amount"       => In8_Customers_Sales_Sync::getSubscriptionProductPrice( $product ),
						"custrecord_in8subs_item_item"         => $internal_id,
						"custrecord_in8subs_item_item_display" => $product->get_title(),
						"custrecord_in8subs_item_pricelevel"   => '',
						"custrecord_in8subs_item_quantity"     => $qty,
					];
					if ( isset( $item['delete'] ) ) {
						$data['delete'] = $item['delete'];
					}
					$data['id'] = $item['id'];
					$_items[]   = $data;
				}
				
				$subscription_items_request  = [
					'recordType'                                  => 'customrecord_in8subs_subscriptions',
					'internalId'                                  => $id,
					'custrecord_in8subs_active'                   => $custrecord_in8subs_active,
					'custrecord_in8subs_bill_order'               => $custrecord_in8subs_bill_order,
					'custrecord_in8subs_billing_address'          => $billingaddress,
//					'custrecord_in8subs_endofcontract' => $custrecord_in8subs_endofcontract,
					'custrecord_in8subs_fulfill_order'            => $custrecord_in8subs_fulfill_order,
					'custrecord_in8subs_get_auth'                 => $custrecord_in8subs_get_auth,
					'custrecord_in8subs_nextpayment'              => $nextPayment,
					'custrecord_in8subs_parent'                   => $customer_internal_id,
					'custrecord_in8subs_payment_method'           => $paymentmethod_id,
					'custrecord_in8subs_recurrence'               => $recurrence_id,
					'custrecord_in8subs_shipping_address'         => $shippingaddress,
					'custrecord_in8subs_startofcontract'          => $start,
					'recmachcustrecord_in8subs_item_subscription' => $_items,
				];
				
				// do not update startofcontract if the subscription is already created
				if( !empty($id) ){
					unset($subscription_items_request['custrecord_in8subs_startofcontract']);
				}
				
				if( !empty($credit_card_token) ){
					$subscription_items_request['custrecord_in8_sub_creditcard_token'] = $credit_card_token;
				}
				$netsuite_connection         = new NetsuiteConnection( $id, 'subscription-items' );
				$subscription_items_request  = apply_filters( 'in8_so_sync_subscription_items_request_data', $subscription_items_request );
				$subscription_items_response = $netsuite_connection->doRequest( $subscription_items_request );
				if ( empty( $subscription_response['data'] ) || $subscription_response['success'] != 1 ) {
					$netsuite_connection->processError( "Failed saving the items for the subscription with id #{$id}." );
					
					return;
				} else {
					$subscriptions_list = $subscription_items_response['subscriptions'];
					if ( ! empty( $user_id ) ) {
						update_user_meta( $user_id, 'subscriptions', $subscriptions_list );
					}
				}
			}
		}
		
		/**
		 * @param WC_Product $product
		 *
		 * @return mixed
		 */
		public static function getSubscriptionProductPrice( $product ) {
			$auto_delivery_price = $product->get_meta( 'auto_delivery_price' );
			
			return ! empty( $auto_delivery_price ) ? $auto_delivery_price : $product->get_price();
		}
		
		/**
		 * @param $address_data
		 * @param $customer_data
		 *
		 * @return array
		 */
		public function get_customer_data_with_address( $address_id, $address_data, $customer_data ) {
			$address_type = $address_data['type'];
			$ad2          = isset( $address_data['billing_address_1'] ) ? 'bill' : 'ship';
			$default      = ( isset( $address_data[ $address_type . '_is_default_address' ] ) && $address_data[ $address_type . '_is_default_address' ] == 1 ) ? 'T' : 'F';
			
			$customer_data = array_merge( $customer_data, [
				$ad2 . 'attention'                             => $address_data['address_internal_name'],
				$ad2 . 'addressee'                             => trim( $address_data[ $address_type . 'first_name' ] . ' ' . $address_data[ $address_type . 'last_name' ] ),
				$ad2 . 'addrphone'                             => $address_data[ $address_type . '_phone' ],
				$ad2 . 'addr1'                                 => $address_data[ $address_type . '_address_1' ],
				$ad2 . 'addr2'                                 => $address_data[ $address_type . '_address_2' ],
				$ad2 . 'city'                                  => $address_data[ $address_type . '_city' ],
				$ad2 . 'state'                                 => $address_data[ $address_type . '_state' ],
				$ad2 . 'zip'                                   => $address_data[ $address_type . '_postcode' ],
				$ad2 . 'label'                                 => $address_data['address_internal_name'],
				$ad2 . 'addresscustom'                         => 'custrecord_in8_' . $ad2 . '_multi_address_id',
				'custrecord_in8_' . $ad2 . '_multi_address_id' => $address_id,
//				'default' . $address_type => $default,
			] );
			
			return $customer_data;
		}
		
		public function get_address_by_id( $user_id, $address_id, $order_id = '' ) {
			$result = [];
			
			
			if ( $address_id === "last_used_billing" || $address_id === "last_used_shipping" || $address_id === "checkout_data" ) {
				/* $prefix = $address_id === 'last_used_shipping' || $address_id === "checkout_data" ? 'shipping' : 'billing';
				$customer = $address_id === "checkout_data" && isset($_POST) ? $_POST : get_user_meta( $user_id); */
				$prefix   = $address_id === 'last_used_shipping' ? 'shipping' : 'billing';
				$customer = get_user_meta( $user_id );
				
				if ( $address_id === "checkout_data" ) {
					$prefix    = isset( $_POST['ship_to_different_address'] ) ? 'shipping' : 'billing';
					$_customer = get_post_meta( $order_id );
					$customer  = [];
					foreach ( $_customer as $key => $value ) {
						$customer[ trim( $key, '_' ) ] = $value;
					}
				}
				
				//new method
				$result = [ 'type' => $prefix ];
				foreach ( (array) $customer as $meta_field_name => $meta_field_value ) {
					if ( isset( $meta_field_value[0] ) && strpos( $meta_field_name, $prefix . "_" ) !== FALSE ) {
						/* wcmca_var_dump($meta_field_name);
						wcmca_var_dump($meta_field_value); */
						$result[ $meta_field_name ] = stripslashes( is_array( $meta_field_value ) ? $meta_field_value[0] : $meta_field_value );
					}
				}
			} else {
				$result = get_user_meta( $user_id, '_wcmca_additional_addresses', TRUE );
				$result = is_array( $result ) && isset( $result[ $address_id ] ) ? $result[ $address_id ] : [];
			}
			
			return $result;
		}
		
		public function get_payment_method_internal_id( $payment_method ) {
			$plugin_options      = In8_Customer::getOptions();
			$plugin_settings_map = [];
			
			if ( ! empty( $plugin_options['mapping_key'] ) && ! empty( $plugin_options['mapping_value'] ) ) {
				$plugin_settings_map = array_combine( $plugin_options['mapping_key'], $plugin_options['mapping_value'] );
				$plugin_settings_map = array_change_key_case( $plugin_settings_map );
			}
			
			$pay_method = 'paymentmethod_' . strtolower( $payment_method );
			if ( ! empty( $plugin_settings_map[ $pay_method ] ) ) {
				return $plugin_settings_map[ $pay_method ];
			}
			
			return FALSE;
		}
		
		public function getSubscriptionRecurrenceOptions() {
			$recurrence_options = [
				'single event' => 5,
				'monthly'      => 3,
				'annual'       => 4,
				'3 months'     => 6,
				'6 months'     => 7,
			];
			$settings           = In8_SalesOrder::getOptions( 2 );
			if ( ! empty( $settings['subscription_recurrence_options'] ) ) {
				$array   = [];
				$options = explode( ',', $settings['subscription_recurrence_options'] );
				if ( is_array( $options ) && count( $options ) >= 1 ) {
					foreach ( $options as $option ) {
						$optionArray = explode( '=>', $option );
						if ( count( $optionArray ) == 2 ) {
							$array[ $optionArray[0] ] = $optionArray[1];
						}
					}
					$recurrence_options = $array;
				}
			}
			
			return $recurrence_options;
		}
		
		public function get_customer_internal_id( $order_id, $user_id = FALSE ) {
			$id = FALSE;
			if ( ! empty( $user_id ) ) {
				$id = get_user_meta( $user_id, 'customerinternalid', TRUE );
			}
			
			if ( empty( $id ) && ! empty( $order_id ) ) {
				$id = get_post_meta( $order_id, 'customerinternalid', TRUE );
			}
			
			return $id;
		}
		
		
		public function handleFailedOrUnsyncedOrders() {
			$co_options = In8_SalesOrder::getOptions( 2 );
			if ( isset( $co_options['sync_failed_orders_cron'] ) && $co_options['sync_failed_orders_cron'] == 'on' ) {
				add_action( 'init', [ $this, 'handleFailedOrUnsyncedOrders_cron' ] );
				add_action( 'in8_netsuite_process_failedOrUnsyncedOrders', [
					$this,
					'handleFailedOrUnsyncedOrders_run_cron',
				] );
			}
			add_action( 'init', function () {
				if ( isset( $_REQUEST['in8_handleFailedOrUnsyncedOrders'] ) ) {
					$this->handleFailedOrUnsyncedOrders_run_cron();
					exit;
				}
			} );
		}
		
		public function handleFailedOrUnsyncedOrders_cron() {
			if ( ! wp_next_scheduled( 'in8_netsuite_process_failedOrUnsyncedOrders' ) ) {
				wp_schedule_event( time(), apply_filters( 'in8_so_failed_orders_cron_schedule', 'hourly' ), 'in8_netsuite_process_failedOrUnsyncedOrders' );
			}
		}
		
		public function handleFailedOrUnsyncedOrders_run_cron() {
			global $wpdb;
			$interval = date( 'Y-m-d h:i:s', time() - 3600 * 24 * 7 );
			$interval = apply_filters('in8_so_failed_orders_cron_interval', $interval);
			
			$post_status = [ 'wc-processing' ];
			
			$options = In8_SalesOrder::getOptions( 2 );
			if ( isset( $options['sync_on_hold'] ) && $options['sync_on_hold'] == 'on' ) {
				$post_status[] = 'wc-on-hold';
			}
			if ( isset( $options['sync_cancelled'] ) && $options['sync_cancelled'] == 'on' ) {
				$post_status[] = 'wc-cancelled';
			}
			if ( isset( $options['sync_on_payment_complete'] ) && $options['sync_on_payment_complete'] == 'on' ) {
				$post_status[] = 'wc-completed';
			}
			$post_status = apply_filters( 'in8_so_failed_orders_cron_order_status', $post_status );
			
			$orders = $wpdb->get_results( "SELECT * FROM {$wpdb->posts} LEFT JOIN (SELECT meta_value, post_id FROM {$wpdb->postmeta} WHERE meta_key='internalid') as missing_internal_id ON missing_internal_id.post_id={$wpdb->posts}.ID WHERE post_type='shop_order' AND post_status IN ('" . implode( "','", $post_status ) . "') AND post_date>'{$interval}' AND (missing_internal_id.meta_value='' OR missing_internal_id.meta_value IS NULL)" );
			
			$orders = apply_filters('in8_so_failed_orders_cron_orders', $orders, $post_status, $interval);
			
			$interval_orders = $orders;
			$email_orders    = $orders;
			
			
			// check if order has already tried to be synced 3 times
			foreach ( $interval_orders as $key => $order ) {
				$tester = new In8_Sync_Tester( wc_get_order( $order->ID ), 'in8_failed_syncing' );
				if ( (integer) $tester->val() >= 4 || $tester->val() === 'Failed 3 times, has to be reviewed.' ) {
					// make the value more readable
					if ( (integer) $tester->val() >= 4 ) {
						$tester->set( 'Failed 3 times, has to be reviewed.' );
					}
					
					unset( $interval_orders[ $key ] );
				}
				if ( $tester->val() !== 'Failed 3 times, has to be reviewed.' ) {
					unset( $email_orders[ $key ] );
				}
				$in8_failed_syncing_notification = get_post_meta( $order->ID, 'in8_failed_syncing_notification', TRUE );
				if ( ! empty( $in8_failed_syncing_notification ) ) {
					unset( $email_orders[ $key ] );
				}
				
				// don't try to sync if if we have it set not to
				if( $this->doNotSync($order->ID,'order') ){
					unset($interval_orders[$key]);
				}
			}
			
			
			// send email notification
			echo "<pre>";
			print_r( 'email orders:' );
			echo "</pre>";
			echo "<pre>";
			print_r( $email_orders );
			echo "</pre>";
			if ( ! empty( $email_orders ) ) {
				$emails_settings = IN8_Settings::getTabSettings( "in8_tab_email", 'unsynced_orders' );
				$orders_message  = array_map( function ( $item ) {
					return "<a href='" . admin_url( "post.php?post={$item->ID}&action=edit" ) . "'>{$item->ID}</a>";
				}, $email_orders );
				$orders_message  = implode( ', ', $orders_message );
				$message         = "List of orders that failed syncing: {$orders_message}";
				if ( ! empty( $emails_settings['unsynced_order_emails'] ) ) {
					wp_mail( explode( ',', $emails_settings['unsynced_order_emails'] ), "IN8 : Unsynced order list for " . get_site_url(), $message, [ 'Content-Type: text/html; charset=UTF-8' ] );
				}
				// stop next email
				foreach ( $email_orders as $key => $order ) {
					update_post_meta( $order->ID, 'in8_failed_syncing_notification', 1 );
				}
			}
			
			echo "<pre>";
			print_r( 'sync orders:' );
			echo "</pre>";
			echo "<pre>";
			print_r( $interval_orders );
			echo "</pre>";
			if ( ! empty( $interval_orders ) ) {
				// sync orders to ns
				foreach ( $interval_orders as $key => $order ) {
					$this->newOrderCreated( $order->ID, FALSE, TRUE );
					$this->sendOrderToNetsuite( $order->ID, 0, FALSE, TRUE );
					
					$tester = new In8_Sync_Tester( wc_get_order( $order->ID ), 'in8_failed_syncing' );
					$tester->increment();
					
					// remove the
					if ( (integer) $tester->val() != 3 ) {
						unset( $interval_orders[ $key ] );
					}
				}
			}
		}
		
		
		/**
		 * @param WC_Order $order
		 *
		 * @return bool
		 */
		public function isCyberSourceTokenizationOn( $order ) {
			return In8_Customer::testFeature( 'enable_cybersource_tokenization' ) && ($order->get_payment_method() == 'cybersource' || $order->get_payment_method() == 'cybersource_credit_card');
		}
		
		public function set_customer_internal_id( $id ) {
			$this->customer_internal_id = $id;
		}
		public function set_order_id( $id ) {
			$this->order_id = $id;
		}
		public function set_user_id( $id ) {
			$this->user_id = $id;
		}
	}
	
	$In8_Customers_Sales_Sync = In8_Customers_Sales_Sync::getInstance();
	
	// TODO - @Robert - why do we need this?
	add_filter( 'in8sync_so_order_sync_customer', function ( $a, $order_id ) {
		return FALSE;
	}, 10, 2 );